"""Thermal modeling tools.

It models conductivity and air layers, but also radiative and convective surface phenomena. Thermal bridges can be added as well as air infiltration. The module yields thermal resistances of walls.

Author: stephane.ploix@grenoble-inp.fr
"""

import warnings
from numpy import interp
import scipy.constants


class Composition:
    """A composition is composed of successive layers related to a unit surface.

    Physical values can be added to the data on top of the class.
    """

    _conductivities = {'concrete': 1, 'glass': 1, 'wood': .2, 'plaster': .35, 'foam': .05, 'polystyrene': .039, 'brick': 0.84}
    _his = {'vertical': 7.69, 'ascending': 10, 'descending': 5.88}
    _he_wind = 5.7
    _he_constant = 11.4
    _emissivities = {'usual': .93, 'glass': .94, 'wood': .95, 'plaster': .98, 'concrete': 0.85, 'polystyrene': .97, 'brick:': .93}
    _thicknesses = (0, 5e-3, 7e-3, 10e-3, 15e-3, 25e-3, 30e-3)
    _thermal_resistances = (0, 0.11, 0.13, 0.15, 0.17, 0.18, 0.18)
    _positions = ['horizontal', 'vertical']
    volumic_masses = {'concrete': 2200, 'air': 1.204, 'wood': 350, 'glass': 2530}
    specific_heats = {'concrete': 880, 'air': 710, 'wood': 2000, 'glass': 720}

    @classmethod
    def _hi_unit(cls, position='vertical'):
        """Retirn indoor convective transmission coefficient.

        :param position: 'vertical' or 'horizontal'
        :type position: string
        :return:  the hi coefficient in W/m2.K
        :rtype: float
        """
        return cls._his[position]

    @classmethod
    def _he_unit(cls, wind_speed_in_ms=2.4):
        """
        Return outdoor convective transmission coefficient.

        :param wind_speed_in_ms: wind speed in m/s, default is to 2.4m/S
        :type wind_speed_in_ms: float
        :return: the coefficient in W/m2.K
        :rtype: float
        """
        return cls._he_wind * wind_speed_in_ms + cls._he_constant

    @classmethod
    def _hr_unit(cls, material, average_temperature_celcius=20):
        """Return radiative transmission coefficient.

        :param material: one string among 'usual', 'glass', 'wood', 'plaster', 'concrete' or 'polystyrene'
        :type material: str
        :param average_temperature_celcius: the temperature for which the coefficient is calculated
        :return: the coefficient in W/m2.K
        :rtype: float
        """
        if material  not in cls._emissivities:
            material = 'usual'
            warnings.warn('material '+material+'has been replaced by usual')
        return 4 * scipy.constants.sigma * cls._emissivities[material] * (average_temperature_celcius + 273.15) ** 3

    @classmethod
    def Rsout_unit(cls, material, average_temperature_celsius, wind_speed_is_m_per_sec):
        """Return outdoor convective and radiative transmission coefficient.

        :param material: one string among 'usual', 'glass', 'wood', 'plaster', 'concrete' or 'polystyrene'
        :type material: str
        :param average_temperature_celsius: the temperature for which the coefficient is calculated
        :type average_temperature_celsius: float
        :param wind_speed_is_m_per_sec: wind speed in m/s
        :type wind_speed_is_m_per_sec: wind spped on site
        :return: the coefficient in W/m2.K
        :rtype: float
        """
        return 1 / (cls._he_unit(wind_speed_is_m_per_sec) + Composition._hr_unit(material, average_temperature_celsius))

    @classmethod
    def Rsin_unit_vertical(cls, material, average_temperature_celsius=20):
        """Indoor convective and radiative transmission coefficient for a vertical surface.

        :param material: one string among 'usual', 'glass', 'wood', 'plaster', 'concrete' or 'polystyrene'
        :type material: str
        :param average_temperature_celsius: the temperature for which the coefficient is calculated
        :type average_temperature_celsius: float
        :return: the coefficient in W/m2.K
        :rtype: float
        """
        return 1 / (Composition._hi_unit('vertical') + Composition._hr_unit(material, average_temperature_celsius))

    @classmethod
    def Rsin_unit_ascending(cls, material='usual', average_temperature_celsius=20):
        """Compute an indoor air limit surface: convective and radiative transmission coefficient for a horizontal surface with ascending heat flow.

        :param material: one string among 'usual', 'glass', 'wood', 'plaster', 'concrete' or 'polystyrene'
        :type material: string
        :param average_temperature_celsius: the temperature for which the coefficient is calculated
        :type average_temperature_celsius: average magnitude temperature i celdius, default to 20.
        :return: the coefficient in W/m2.
        :rtype: float
        """
        return 1 / (Composition._hi_unit('ascending') + Composition._hr_unit(material, average_temperature_celsius))

    @classmethod
    def Rsin_unit_descending(cls, material='usual', average_temperature_celsius=20):
        """Compute: Indoor convective and radiative transmission coefficient for a horizontal surface with descending heat flow.

        :param material: one string among 'usual', 'glass', 'wood', 'plaster', 'concrete' or 'polystyrene'
        :type material: str
        :param average_temperature_celsius: the temperature for which the coefficient is calculated, default is to 20
        :type average_temperature_celsius: float
        :return: the coefficient in W/m2.K
        :rtype: float
        """
        return 1 / (Composition._hi_unit('descending') + Composition._hr_unit(material, average_temperature_celsius))

    @classmethod
    def R_air_layer_unit_resistance(cls, thickness, position, material1, material2, average_temperature_celsius=20):
        """Return the hermal resistance of an air layer depending of its position.

        :param thickness: thickness of the air layer
        :type thickness: float
        :param position: 'horizontal' or 'vertical'
        :type position: float
        :param material1: one string among 'usual', 'glass', 'wood', 'plaster', 'concrete' or 'polystyrene'
        :type material1: str
        :param material2: one string among 'usual', 'glass', 'wood', 'plaster', 'concrete' or 'polystyrene'
        :type material2: str
        :param average_temperature_celsius: the temperature for which the coefficient is calculated, default is to 20
        :type average_temperature_celsius: float
        :return: thermal resistance in K.m2/W
        :rtype: float
        """
        if thickness <= cls._thicknesses[-1]:
            return interp(thickness, cls._thicknesses, cls._thermal_resistances, left=0, right=cls._thermal_resistances[-1])
        else:
            if position == 'vertical':
                return cls.Rsin_unit_vertical(material1, average_temperature_celsius) + cls.Rsin_unit_vertical(material2, average_temperature_celsius)
            else:
                return cls.Rsin_unit_descending(material2, average_temperature_celsius) + cls.Rsin_unit_ascending(material1, average_temperature_celsius)

    @classmethod
    def R_unit_conduction(cls, thickness, material):
        """Compute the conductive resistance of an layer depending of its thickness.

        :param thickness: thickness of the layer
        :type thickness: float
        :param material: one string among 'usual', 'glass', 'wood', 'plaster', 'concrete' or 'polystyrene'
        :type material: str
        :return: thermal resistance in K.m2/W
        :rtype: float
        """
        return thickness / cls._conductivities[material]

    def _R_unit_added(self, layer_material):
        """Compute unit resistance of a layer added to a composition under construction.

        :param layer_material: one string among 'usual', 'glass', 'wood', 'plaster', 'concrete' or 'polystyrene'
        :type layer_material: str
        :return: unit resistance
        :rtype: float
        """
        if not self.first_layer_indoor:
            return Composition.Rsout_unit(material=layer_material, average_temperature_celsius=self.outdoor_average_temperature, wind_speed_is_m_per_sec=self.wind_speed_is_m_per_sec)
        else:
            if self.position == 'vertical':
                return Composition.Rsin_unit_vertical(material=layer_material, average_temperature_celsius=self.indoor_average_temperature)
            elif self.position == 'horizontal':
                if not self.heating_floor:
                    return Composition.Rsin_unit_descending(material=layer_material, average_temperature_celsius=self.indoor_average_temperature)
                else:
                    return Composition.Rsin_unit_ascending(material=layer_material, average_temperature_celsius=self.indoor_average_temperature)
            raise ValueError('Unknown position')

    def __init__(self, first_layer_indoor, last_layer_indoor, position, indoor_average_temperature_in_celsius=20, outdoor_average_temperature_in_celsius=5, wind_speed_is_m_per_sec=2.4, heating_floor=False):
        """Create a composition is composed of successive layers.

        Physical values can be added to the data on top of the class.

        :param first_layer_indoor: True if first layer is indoor, False if it's outdoor and None if it's not in contact with air
        :type first_layer_indoor: bool or None
        :param last_layer_indoor: True if last layer is indoor, False if it's outdoor and None if it's not in contact with air
        :type last_layer_indoor: bool or None
        :param position: can be either 'vertical' or 'horizontal'
        :type position: str
        :param indoor_average_temperature_in_celsius: average indoor temperature used for linearization of radiative heat in Celsius, default is to 20°C
        :type indoor_average_temperature_in_celsius: float
        :param outdoor_average_temperature_in_celsius: average outdoor temperature used for linearization of radiative heat in Celsius, default is to 5°C
        :type outdoor_average_temperature_in_celsius: float
        :param wind_speed_is_m_per_sec (default is 2.4m/s): average wind speed in m/s
        :type wind_speed_is_m_per_sec: float
        :param heating_floor: True if there is a heating floor, False otherwise. Default is to False
        :type heating_floor: bool
        """
        self.position: str = position
        self.first_layer_indoor: bool = first_layer_indoor
        self.last_layer_indoor: bool = last_layer_indoor
        self.indoor_average_temperature: float = indoor_average_temperature_in_celsius
        self.outdoor_average_temperature: float = outdoor_average_temperature_in_celsius
        self.wind_speed_is_m_per_sec: float = wind_speed_is_m_per_sec
        self.heating_floor: bool = heating_floor
        self.layers = list()

    def add_layer(self, material, thickness):
        """Add a layer in the composition. First and last layers cannot be air layers.

        :param material: name of the material
        :type material: str
        :param thickness: thickness of the layer
        :type thickness: float
        """
        self.layers.append({'material': material, 'thickness': thickness})

    @property
    def R(self):
        """Compute the thermal resistance for the composition in W.m2/K.

        :return: unit resistance value (for 1m2)
        :rtype: float
        """
        _resistance = 0
        if self.first_layer_indoor is not None:
            if self.layers[0]['material'] == 'air' or self.layers[-1]['material'] == 'air':
                raise ValueError('an external layer cannot be made of air')
            _resistance += self._R_unit_added(self.layers[0]['material'])
        for l in range(len(self.layers)):
            if self.layers[l] is not None:
                if self.layers[l]['material'] == 'air':
                    radiative_resistance = (1 / Composition._hr_unit(self.layers[l-1]['material'], self.indoor_average_temperature)) + 1 / Composition._hr_unit(self.layers[l+1]['material'], self.indoor_average_temperature)
                    position = self.position
                    if self.position == 'horizontal' and self.heating_floor:
                        position = 'ascending'
                    elif self.position == 'horizontal' and not self.heating_floor:
                        position = 'descending'
                    air_layer_resistance = Composition.R_air_layer_unit_resistance(self.layers[l]['thickness'], position, self.layers[l-1]['material'], self.layers[l+1]['material'])
                    _resistance += 1 / (1 / air_layer_resistance + 1 / radiative_resistance)
                else:
                    _resistance += Composition.R_unit_conduction(self.layers[l]['thickness'], self.layers[l]['material'])
        if self.last_layer_indoor is not None:
            _resistance += self._R_unit_added(self.layers[-1]['material'])
        return _resistance

    @property
    def U(self):
        """Compute thermal transmission coefficient in W/m2.K (1/R of the unit resistance).

        :return: thermal transmission coefficient in W/m2.K
        :rtype: float
        """
        return 1/self.R

    def __str__(self):
        """Return a descriptive string.

        :return: string representative of the composition
        """
        string = 'U=%.4fW/K.m2 (R_unit=%.4fK.m2/W) composed of ' % (self.U, self.R)
        if self.first_layer_indoor is None:
            string += 'contact:'
        elif self.first_layer_indoor:
            string += 'indoor(%s):' % self.position
        else:
            string += 'outdoor(%s):' % self.position
        for layer in self.layers:
            string += '%s(%.3fm):' % (layer['material'], layer['thickness'])
        if self.last_layer_indoor is None:
            string += 'contact\n'
        elif self.last_layer_indoor:
            string += 'indoor(%s)\n' % self.position
        else:
            string += 'outdoor(%s)\n' % self.position
        return string


class Wall:
    """A wall is a set of compositions and thermal bridges of different dimensions (surface for compositions and length for thermal bridges) but also an infiltration varying from a minimum to a maximum value in m3/s."""

    def __init__(self, name):
        """Create a wall.

        :param name: name of the wall.
        :type name: str
        """
        self.name = name
        self.compositions = list()
        self.surfaces = list()
        self.bridges = list()
        self.lengths = list()
        self._infiltration_air_flow = 0
        self._max_opening_air_flow = 0

    def add_composition(self, composition, surface):
        """Add a part of wall in parallel with others defined by a composition and a surface.

        :param composition: composition of the part of wall
        :type composition: Composition
        :param surface: surface of the part of wall
        :type surface: float
        """
        self.compositions.append(composition)
        self.surfaces.append(surface)

    def add_bridge(self, psi, length):
        """Add a thermal bridge.

        :param psi: linear transmission coefficient in W/m.K
        :type psi: float
        :param length: length of the thermal bridge
        :type length: float
        """
        self.bridges.append(psi)
        self.lengths.append(length)

    def add_infiltration(self, minimum_infiltration):
        """Add a minimum infiltration air flow through the wall.

        :param minimum_infiltration: minimum infiltration air flow in m3/s
        :type minimum_infiltration: float
        """
        self._infiltration_air_flow = minimum_infiltration

    def add_max_opening_air_flow(self, maximum_infiltration):
        """Add a maximum infiltration air flow reached when opening ration is equal to 1.

        :param maximum_infiltration: maximum infiltration air flow in m3/s
        :type maximum_infiltration: float
        """
        self._max_opening_air_flow = maximum_infiltration

    def air_volumic_thermal_transmission(self, opening_ratio=0):
        """Return heat transmitted by volumetric exchange for an intermediate value of infiltration.

        :param opening_ratio: value between 0 and 1 changing the infiltration air flow from minimum to maximum, default is to 0.
        :type opening_ratio: float
        :return: heat transmitted by volumetric exchange in W/K
        :rtype: float
        """
        Q = self._infiltration_air_flow + opening_ratio * (self._max_opening_air_flow - self._infiltration_air_flow)
        return Composition.volumic_masses['air'] * Composition.specific_heats['air'] * Q

    def US(self, opening_ratio=0):
        """Compute the total heat loss of the wall in W/K.

        :param opening_ratio: value between 0 and 1 changing the infiltration aor flow from minimum to maximum
        :type opening_ratio: float
        :return: total heat loss of the wall in W/K
        :rtype: float
        """
        thermal_transmission_coefficient = 0
        for i in range(len(self.compositions)):
            thermal_transmission_coefficient += self.compositions[i].U * self.surfaces[i]
        for i in range(len(self.bridges)):
            thermal_transmission_coefficient += self.bridges[i] * self.lengths[i]
        thermal_transmission_coefficient += self.air_volumic_thermal_transmission(opening_ratio)
        return thermal_transmission_coefficient

    def R(self, opening_ratio=0):
        """Compute the total thermal resistance of the wall.

        :param opening_ratio: value between 0 and 1 changing the infiltration air flow from minimum to maximum
        :type opening_ratio: float
        :return: total thermal resistance of the wall in K/W
        :rtype: float
        """
        return 1 / self.US(opening_ratio)

    def __str__(self):
        """Return a descriptive string.

        :return: string representative of the wall
        """
        string = '+ Wall % s,' % self.name
        if self.US() == self.US(opening_ratio=1):
            string += 'US=%fW/K (R=%fK/W)\n' % (self.US(), self.R())
        else:
            string += 'US=%f>%fW/K (R=%f<%fK/W)\n' % (self.US(), self.US(1), self.R(), self.R(1))
        for i in range(len(self.compositions)):
            string += 'Composition(%.2fm2):' % (self.surfaces[i])
            string += self.compositions[i].__str__()
        for i in range(len(self.bridges)):
            string += 'Thermal bridge (length=%fm): psiL=%fW/K\n' % (self.lengths[i], self.lengths[i]*self.bridges[i])
        if self.air_volumic_thermal_transmission() == self.air_volumic_thermal_transmission(opening_ratio=1) == 0:
            return string + '\n'
        string += 'Infiltration loss: U=%fW/K' % self.air_volumic_thermal_transmission()
        if self.air_volumic_thermal_transmission(opening_ratio=1)!=0:
            string += '>%fW/K' % self.air_volumic_thermal_transmission(opening_ratio=1)
        return string + '\n'



"""Set of tools to model solar masks for a building.

Author: stephane.ploix@grenoble-inp.fr
"""

from abc import ABC
from buildingenergy import solar
import pyexcelerate
import matplotlib.pyplot
from math import pi

class Zone(ABC):
    """Abstract class standing for a single zone building. It can be used once specialized."""

    _azimut_min_max_in_rad: tuple = (-pi, pi)
    _altitude_min_max_in_rad: tuple = (0, pi / 2)
    _mask_plot_resolution: int = 20

    @staticmethod
    def set_max_plot_resolution(number_of_sample_per_axis):
        """Redefine the default grid max_plot resolution for plotting the zone.

        :param number_of_sample_per_axis: number of Solar Position per axis
        :return: None
        """
        Zone._mask_plot_resolution = number_of_sample_per_axis

    def __init__(self):
        """Create a zone."""
        pass

    def __contains__(self, solar_position):
        """Determine whether a solar_position of the sun in the sky defined by 2 angles (azimut and altitude) is contained by the zone or not.

        :param solar_position: solar solar_position in the sky
        :type solar_position: SolarPosition
        :return: True if contained, False otherwise
        :rtype: bool
        """
        raise NotImplementedError

    def plot(self, name=''):
        """Plot the mask according to the specified max_plot_resolution and print a description of the zone.

        :param name: file_name of the plot, default to ''
        :return: the zone
        """
        matplotlib.pyplot.figure()
        azimuts = [Zone._azimut_min_max_in_rad[0] + i * (Zone._azimut_min_max_in_rad[1] - Zone._azimut_min_max_in_rad[0]) / (Zone._mask_plot_resolution - 1) for i in range(Zone._mask_plot_resolution)]
        altitudes = [-Zone._altitude_min_max_in_rad[0] + i * (Zone._altitude_min_max_in_rad[1] - Zone._altitude_min_max_in_rad[0]) / (Zone._mask_plot_resolution - 1) for i in range(Zone._mask_plot_resolution)]
        axes = matplotlib.pyplot.gca()
        axes.set_xlim((180 / pi * Zone._azimut_min_max_in_rad[0], 180 / pi * Zone._azimut_min_max_in_rad[1]))
        axes.set_ylim((-180 / pi * Zone._altitude_min_max_in_rad[0], 180 / pi * Zone._altitude_min_max_in_rad[1]))
        for azimut in azimuts:
            for altitude in altitudes:
                if SolarPosition(azimut * 180 / pi, altitude * 180 / pi) in self:
                    matplotlib.pyplot.scatter(180/pi*azimut, 180/pi*altitude, c=0)
        matplotlib.pyplot.title(name)
        print('- %s' % name)
        print(self.__str__())
        return self


class SolarPosition:
    """Contain azimut and altitude angles of the sun for mask processing."""

    def __init__(self, azimut_in_deg, altitude_in_deg):
        """Create a solar position with azimut (south to west directed) and altitude (south to zenith directed) angles of the sun for mask processing.

        :param azimut_in_deg: solar angle with the south in degree, counted with trigonometric direction i.e. the azimut (0=South, 90=West, 180=North, -90=East)
        :type azimur_in_deg: float
        :param altitude_in_deg: zenital solar angle with the horizontal in degree i.e. the altitude (0=horizontal, 90=zenith)
        :type altitude_in_deg: float
        """
        self.azimut: float = azimut_in_deg/180*pi
        self.altitude: float = altitude_in_deg/180*pi

    @property
    def normalized_azimut_in_rad(self):
        """Return normalized solar azimut to belong to (-pi,pi) rad.

        :return: normalized azimut
        :rtype: float
        """
        return SolarPosition._normalize(self.azimut)

    @property
    def normalized_altitude_in_rad(self):
        """Return normalized solar altitude to belong to (-pi,pi) rad.

        :return: normalized altitude
        """
        return SolarPosition._normalize(self.altitude)

    @staticmethod
    def _normalize(angle):
        """Return a normalization of an angle for its value belongs to (-pi,pi) rad.

        :param angle: angle to be normalized
        :type angle: float
        :return: normalized angle in rad.
        :rtype: float
        """
        if angle % (2 * pi) > pi:
            angle = angle % (2 * pi) - 2 * pi
        if angle % (2 * pi) < -pi:
            angle = angle % (2 * pi) + 2 * pi
        return angle

    def __str__(self):
        """Return a description of the solar position.

        :return: string with normalized azimut and altitude angles in degree
        :rtype: str
        """
        return '(AZ:%frad,AL:%frad)' % (self.normalized_azimut_in_rad / pi * 180, self.normalized_altitude_in_rad / pi * 180)


class _BasicZone(Zone):
    """Abstract class that specifies basic zones."""

    def __init__(self):
        Zone.__init__(self)

    def __contains__(self, solar_position):
        raise NotImplemented


class _VerticalZone(_BasicZone):
    """A half vertical plan mask."""

    def __init__(self, solar_position, left):
        """Create a half vertical plan mask.

        :param solar_position: a solar position whose azimut is used to define the vertical half plan
        :type solar_position: SolarPosition
        :param left: define whether the left hand side should be considered (True) or the right hand side (False)
        :type left: bool
        """
        _BasicZone.__init__(self)
        self.position = solar_position
        self.left = left

    def __contains__(self, solar_position):
        """Check weither a position belong too the current half plan.

        :param solar_position:
        :type SolarPosition:
        :return: True if in the mask, False otherwise
        :rtype: bool
        """
        if self.left:
            if solar_position.normalized_azimut_in_rad <= self.position.normalized_azimut_in_rad:
                return True
        else:  # right
            if solar_position.normalized_azimut_in_rad >= self.position.normalized_azimut_in_rad:
                return True
        return False

    def __str__(self):
        """Return a representation of the half plan.

        :return: a descriptive string
        :rtype: str
        """
        if self.left:
            return '(left AZ:%f)' % (self.position.normalized_azimut_in_rad / pi * 180)
        else:
            return '(right AZ:%f)' % (self.position.normalized_azimut_in_rad / pi * 180)


class _HorizontalZone(_BasicZone):
    """A half horizontal plan mask."""

    def __init__(self, solar_position, low):
        """Create a half horizontal plan mask.

        :param solar_position: a solar position whose altitude is used to define the horizontal half plan
        :type solar_position: SolarPosition
        :param low: define whether the lower half plan should be considered (True) or the upper one (False)
        :type low: bool
        """
        _BasicZone.__init__(self)
        self.solar_position = solar_position
        self.low = low

    def __contains__(self, solar_position):
        """Check weither a position belong too the current half plan.

        :param solar_position: a solar position
        :type solar_position: SolarPosition
        :return: True if in the mask, False otherwise
        :rtype: bool
        """
        if solar_position.normalized_altitude_in_rad <= self.solar_position.normalized_altitude_in_rad:
            return self.low
        else:
            return not self.low

    def __str__(self):
        """Return a description of the half horizontal zone mask.

        :return: a string depicting the mask
        :rtype: str
        """
        if self.low:
            return '(lower AL:%f)' % (self.solar_position.normalized_altitude_in_rad / pi * 180)
        else:
            return '(upper AL:%f)' % (self.solar_position.normalized_altitude_in_rad / pi * 180)


class ComplexZone(Zone):
    """A complex zone is a set of masks with an intersection or a union in between."""

    def __init__(self, intersect: bool=True):
        """Create a complex zone.

        :param intersect: intersection if True, union otherwise
        :type intersect: bool
        """
        Zone.__init__(self)
        self.zones = list()
        self.intersect: bool = intersect
        self._inverse = False

    def inverse(self):
        """Invert the zone.

        :return: the inverted zone
        :rtype: ComplexZone
        """
        self._inverse = not self._inverse
        return self

    def union(self):
        """Transform the zone into a union.

        :return: the zone
        :rtype: ComplexZone
        """
        self.intersect: bool = False
        return self

    def intersection(self):
        """Transform the zone into an intersection.

        :return: the zone
        :rtype: ComplexZone
        """
        self.intersect: bool = True
        return self

    def add_vertical_strip_zone(self, left_solar_position, right_solar_position):
        """Create and add to the zone a vertical strip zone defined by the azimuts of two solar positions. The order matters: if the azimut of the right solar position is lower than the one of left solar position, it is understood than the strip include azimut 180°.

        :param left_solar_position: solar position specifying the left hand side azimut
        :type left_solar_position: SolarPosition
        :param right_solar_position: solar position specifying the right hand side azimut
        :type right_solar_position: SolarPosition
        :return: the zone
        :rtype: ComplexZone
        """
        if left_solar_position.normalized_azimut_in_rad <= right_solar_position.normalized_azimut_in_rad:
            vertical_strip_zone = ComplexZone(intersect=True)
            vertical_strip_zone.add_vertical_zone(left_solar_position, left=False)
            vertical_strip_zone.add_vertical_zone(right_solar_position, left=True)
        else:
            vertical_strip_zone = ComplexZone(intersect=False)
            vertical_strip_zone.add_vertical_zone(left_solar_position, left=False)
            vertical_strip_zone.add_vertical_zone(right_solar_position, left=True)
        self.zones.append(vertical_strip_zone)
        return self

    def add_horizontal_strip_zone(self, lower_solar_position, upper_solar_position):
        """Create and add to the zone a horizontal strip zone defined by the altitudes of two solar positions.

        :param lower_solar_position: solar position specifying the lower altitude
        :type lower_solar_position: SolarPosition
        :param upper_solar_position: solar position specifying the upper altitude
        :type upper_solar_position: SolarPosition
        :return: the zone
        :rtype: ComplexZone
        """
        complex_zone = ComplexZone(intersect=True)
        complex_zone.add_horizontal_zone(lower_solar_position, low=False)
        complex_zone.add_horizontal_zone(upper_solar_position, low=True)
        self.zones.append(complex_zone)
        return self

    def add_horizontal_zone(self, solar_position, low):
        """Create and add to the zone an horizontal half plan zone.

        :param solar_position: the solar position whose altitude specifies the horizontal half plan
        :type solar_position: SolarPosition
        :param low: lower half plan if True, upper otherwise
        :type low: bool
        :return: the zone
        :rtype: ComplexZone
        """
        self.zones.append(_HorizontalZone(solar_position, low=low))
        return self

    def add_vertical_zone(self, solar_position, left):
        """Create and add to the zone a vertical half plan zone.

        :param solar_position: the solar position whose azimut specifies the vertical half plan
        :type solar_position: SolarPosition
        :param left: left side if True, right side otherwise
        :type left: bool
        :return: the zone
        :rtype: ComplexZone
        """
        self.zones.append(_VerticalZone(solar_position, left=left))
        return self

    def add_rectangle_zone(self, lower_left_solar_position, upper_right_solar_position):
        """Create and add to the zone a rectangle defined by two solar positions. The order matters: if the azimut of the upper right solar position is lower than the one of lower left solar position, it is understood than the strip include azimut 180°.

        :param lower_left_solar_position: lower left solar position
        :type lower_left_solar_position: SolarPosition
        :param upper_right_solar_position: upper right solar position
        :type upper_right_solar_position: SolarPosition
        :return: the zone
        :rtype: ComplexZone
        """
        complex_zone = ComplexZone(intersect=True)
        complex_zone.add_vertical_strip_zone(lower_left_solar_position, upper_right_solar_position)
        complex_zone.add_horizontal_strip_zone(lower_left_solar_position, upper_right_solar_position)
        self.zones.append(complex_zone)
        return self

    def add_vertical_window_mask(self, south_relative_direction_in_deg=0, altitude_limitation_in_deg=90, azimut_limitation_in_deg=90):
        """Create and add a mask modeling the obstacles around a vertical window. The default parameters keep only the solar radiation in the south_relative_direction_in_deg +/- 90° in azimut and (0°,90°) in altitude. An upper solar protection reduces the range of possible solar altitudes. It can be modified by reducing altitude_limitation_in_deg. Left and right solar protections reduce the range of possible solar azimuts. It can be modified by asjusting azimut_limitation_in_deg.

        :param south_relative_direction_in_deg: direction perpendicular to the window, default to 0.
        :type south_relative_direction_in_deg: float
        :param altitude_limitation_in_deg: altitude limitation. 0 means no more direct solar radiation and 90° means no limitation on the altitude, default to 90.
        :type altitude_limitation_in_deg: float
        :param azimut_limitation_in_deg: limitation on the visible azimuts. 0 means no more direct solar radiation and 90° stands for no limitation on a plan perpendicular to the south_relative_direction_in_deg, default to 90.
        :type azimut_limitation_in_deg: float
        :return: the zone
        :rtype: ComplexZone
        """
        lower_left_position = SolarPosition(south_relative_direction_in_deg - azimut_limitation_in_deg, -altitude_limitation_in_deg)
        upper_right_position = SolarPosition(south_relative_direction_in_deg + azimut_limitation_in_deg, altitude_limitation_in_deg)
        self.add_rectangle_zone(lower_left_position, upper_right_position)
        return self

    def add_zone(self, zone):
        """Add a zone to the current one.

        :param zone: zone to be added
        :type zone: Zone
        :return: the zone
        :rtype: ComplexZone
        """
        self.zones.append(zone)
        return self

    def __contains__(self, solar_position):
        """Check weither the sun is masked or not.

        :param solar_position: azimut and altitude of the sun
        :type solar_position: SolarPosition
        :return: True if in the mask or False otherwise
        :rtype: bool
        """
        _inside = self.intersect
        for zone in self.zones:
            if self.intersect:
                _inside = _inside and (solar_position in zone)
            else:
                _inside = _inside or (solar_position in zone)
        if self._inverse:
            return not _inside
        return _inside

    def __str__(self):
        """Return a description of the complex zone.

        :return: a string describing the complex zone mask
        :rtype: bool
        """
        if self.intersect:
            string = 'Inter['
        else:
            string = 'Union['
        for i in range(len(self.zones)):
            string += self.zones[i].__str__()
            if i+1 < len(self.zones):
                string += ','
        return '\t' + string + ']'


class WindowMask(ComplexZone):
    """A specilized complex zone useful when mask is rectangular."""

    def __init__(self, azimut: tuple, altitude: tuple):
        """Create a window mask.

        :param azimut: east-west angle clockwise
        :type azumut: tuple[float]
        :param altitude: altitude or elevation angle in degree
        :type altitude: tuple[float]
        """
        super().__init__(intersect=True)
        self.add_rectangle_zone(SolarPosition(min(azimut), min(altitude)), SolarPosition(max(azimut), max(altitude)))


class Building:
    """A class used to obtain the solar gains through the windows of a building."""

    def __init__(self, site_weather_data, solar_mask: ComplexZone = None):
        """Create a parallelepipedic building with different windows with solar masks to estimate the solar gain.

        :param site_weather_data: wheather data
        :type site_weather_data: openweather.SiteWeatherData
        :param solar_mask: distant solar mask used for the whole building. None means no global solar masks
        :type solar_mask: ComplexZone
        """
        self.site_weather_data = site_weather_data
        self._stringdates, self._datetimes = self.site_weather_data.get('stringdate'), self.site_weather_data.get('datetime')
        if solar_mask is None:
            self.solar_mask = ComplexZone()
        else:
            self.solar_mask = solar_mask
        self.windows = dict()
        self.solar_model = solar.SolarModel(self.site_weather_data)

    def add_window(self, name, surface, exposure_in_deg=0, slope_in_deg=90, solar_factor=0.85, window_mask=None):
        """Add a window to the building.

        :param name: name of the window
        :type name: str
        :param surface: surface of the glass
        :type surface: float
        :param exposure_in_deg: exposure of the window (default: 0 = south) i.e. angle between the south direction and the normal to the surface projected in the tangential plan (0° means south directed, -90° east and 90° west
        :type exposure_in_deg: float
        :param slope_in_deg: slope of the window (default: 90 = vertical) i.e. angle between a tangential plan and the surface (0° means facing ground, 90° vertical and 180° facing the sky)
        :type slope_in_deg: float
        :param solar_factor: solar factor between 0 (opaque) and 100 (fully transparent)
        :type solar_factor: float
        :param window_mask: mask only for this window
        :type window_mask: WindowMask
        """
        if window_mask is not None:
            self.solar_mask.add_zone(window_mask)
        slope_in_deg = slope_in_deg - 90
        self.windows[name] = {'surface': surface, 'exposure_in_deg': exposure_in_deg, 'slope_in_deg': slope_in_deg, 'solar_factor': solar_factor}

    @property
    def datetimes(self):
        """Provide dates coming from weather data in a Python format.

        :return: list of dates in datetime.datetime format
        :rtype: list[datetime.datetime]
        """
        return self._datetimes

    @property
    def stringdates(self):
        """Return dates coming from weather data in a string format.

        :return: list of string dates
        :rtype: list[str]
        """
        return self._stringdates

    @property
    def solar_gain(self):
        """Return hourly solar gains coming through the windows.

        :return: list of total solar gains powers in Watt, and a dictionary of the detailed solar gains in Watts coming through each window
        :rtype: tuple[list[float], dict[str,list[float]]]
        """
        phi_total = list()
        phi_windows = dict()
        for window_name in self.window_names:
            phi_windows[window_name] = list()
        for k in range(len(self._datetimes)):
            phi_hour = 0
            current_datetime = self._datetimes[k]
            altitude_in_deg, azimuth_in_deg = self.solar_model.solar_angle(current_datetime)
            current_position = SolarPosition(azimuth_in_deg, altitude_in_deg)
            if altitude_in_deg > 0:
                for window_name in self.windows:
                    _, phi_direct_collected, phi_diffuse, phi_reflected = self.solar_model.solar_irradiation(self.windows[window_name]['exposure_in_deg'], self.windows[window_name]['slope_in_deg'], current_datetime, temperature=self.site_weather_data.get('temperature')[k], humidity=self.site_weather_data.get('humidity')[k], nebulosity_in_percentage=self.site_weather_data.get('cloudiness')[k], pollution=self.site_weather_data.pollution)
                    if self.solar_mask is None or current_position in self.solar_mask:
                        phi_window_hour = self.windows[window_name]['solar_factor'] * self.windows[window_name]['surface'] * (phi_direct_collected + phi_diffuse + phi_reflected)
                    else:
                        phi_window_hour = self.windows[window_name]['solar_factor'] * self.windows[window_name]['surface'] * (phi_diffuse + phi_reflected)
                    phi_hour += phi_window_hour
                    phi_windows[window_name].append(phi_window_hour)
                phi_total.append(phi_hour)
            else:
                phi_total.append(0)
                for window_name in self.windows:
                    phi_windows[window_name].append(0)
        return phi_total, phi_windows

    @property
    def window_names(self):
        """Return all the window names.

        :return: names of the windows
        :rtype: tuple[str]
        """
        return tuple(self.windows.keys())

    def window(self, name):
        """Return the solar gain through a window.

        :param name: name of a window
        :type name: str
        :return: list of solar powers crossing the window glass
        :rtype: list[float]
        """
        _, phi_windows = self.solar_gain
        return phi_windows[name]

    def day_averager(self, vector, average=True):
        """Compute the average or integration hour-time data to get day data.

        :param average: True to compute an average, False for integration
        :type average: bool
        :param vector: vector of values to be downsampled
        :return: list of floating numbers
        """
        current_day = self.datetimes[0].day
        day_integrated_vector = list()
        values = list()
        for k in range(len(self.datetimes)):
            if current_day == self.datetimes[k].day:
                values.append(vector[k])
            else:
                average_value = sum(values)
                if average:
                    average_value = average_value/len(values)
                day_integrated_vector.append(average_value)
                values = list()
            current_day = self.datetimes[k].day
        return day_integrated_vector

    def __len__(self):
        """Return the number of hours in the weather data.

        :return: number of hours in the weather data
        :rtype: int
        """
        return len(self.stringdates)

    def generate_xls(self, file_name='calculations', heat_temperature_reference=18,  cool_temperature_reference=26):
        """Save day degrees and solar gains per window for each day in an xls file.

        :param file_name: file name without extension, default to 'calculation'
        :type file_name: str
        :param temperature_reference: reference temperature for heating day degrees
        :type heat_temperature_reference: float
        :param cool_temperature_reference: reference temperature for cooling day degrees
        :type cool_temperature_reference: float
        """
        stringdate_days, average_temperature_days, min_temperature_days, max_temperature_days, dju_heat_days = self.site_weather_data.day_degrees(temperature_reference=heat_temperature_reference, heat=True)
        _, _, _, _, dju_cool_days = self.site_weather_data.day_degrees(temperature_reference=cool_temperature_reference, heat=False)

        data = [['date'], ['Tout'], ['Tout_min'], ['Tout_max'], ['dju_heat'], ['dju_cool']]
        data[0].extend(stringdate_days)
        data[1].extend(average_temperature_days)
        data[2].extend(min_temperature_days)
        data[3].extend(max_temperature_days)
        data[4].extend(dju_heat_days)
        data[5].extend(dju_cool_days)

        i=6
        for window_name in self.window_names:
            data.append([window_name+'(Wh)'])
            data[i].extend(self.day_averager(self.window(window_name), average=False))
            i += 1

        excel_workbook = pyexcelerate.Workbook()
        excel_worksheet = excel_workbook.new_sheet(file_name, data=list(map(list, zip(*data))))
        excel_workbook.save(file_name + '.xlsx')



"""ARX design helper.

Author: stephane.ploix@g-scop.grenoble-inp.fr
"""

import copy
import math
import matplotlib
import matplotlib.pyplot
import numpy
import numpy.linalg


class LinearRegression:
    """General class to design ARX model matching measurement data."""

    def __init__(self, input_labels, output_label, minimum_input_delay, inputs_maximum_delays, output_maximum_delay, offset=False):
        """Estimate linear regression coefficients for a dataset.

        The meta-parameters for tuning the ARX model structre are:
            offset = True
            minimum_input_delay = 1
            inputs_maximum_delays = [2 3] (if a single int value, 4 for instance, is used, it would be equal to [4 4])
            ouput_maximum_delay = 3
        we can get the following linear regression:
            y{k} = c1 y{k-1} + c2 y{k-2} + c3 y{k-3} (ouput_maximum_delay)
                ...+ c4 u1{k-1} (minimum_input_delay) + c5 u1{k-2} (inputs_maximum_delays[0])
                ...+ c6 u2{k-1} (minimum_input_delay) + c7 u2{k-2} + c8 u2{k-3} (inputs_maximum_delays[1])
                ...+ c9

        :param input_label: input variable names
        :type input_label: list[str]
        :param output_label: output variable name
        :type output_label: list[str]
        :param minimum_input_delay: int standing for the delay_order variable represents the time delay to the system
        :type minimum_input_delay: int
        :param inputs_maximum_delays: int or list standing for the num_orders variable represents the order for the different inputs
        :type inputs_maximum_delays: int or list[int]
        :param output_maximum_delay: int standing for the den_order variable represents the order for the output
        :type outputs_maximum_delays: int or list[int]
        :param offset: will search for an offset constant value if True
        :type offset: bool
        """
        self.__input_labels = input_labels
        self.__number_of_inputs = len(input_labels)
        self.__output_label = output_label
        self.__offset = offset
        if type(inputs_maximum_delays) == int:
            inputs_maximum_delays = [inputs_maximum_delays for i in range(len(input_labels))]
        self.__inputs_maximum_delays = [max(minimum_input_delay, num_order) for num_order in inputs_maximum_delays]
        self.__output_maximum_delay = output_maximum_delay
        self.__input_minimum_delay = minimum_input_delay
        self.__output_delay_parameters: dict[int, float] = dict()
        self.__inputs_delay_parameters: dict[int, float] = [dict() for _ in range(self.__number_of_inputs)]
        self.__offset_value: float = 0
        self.__output_delay_impacts = {delay: 0 for delay in self.__output_delay_parameters}
        self.__inputs_delay_impacts = [{delay: 0 for delay in self.__inputs_delay_parameters[i]} for i in range(self.__number_of_inputs)]
        self.__offset_impact = 0
        self.__total_impact = 0

    def learn(self, list_of_input_values, output_values):
        """Estimate the parameters of the linear regression according to the structure specified in the initializer.

        :param list_of_input_values: list of variable data to be used for input variables where size = number of input variables x number of data
        :type list_of_input_values: list[list[float]]
        :param output_values: list of variable data to be used for output variables where size = number of data
        :type output_values: list[float]
        """
        __number_of_values = len(output_values)
        __Y_matrix, __X_matrix = list(), list()
        for k in range(max(self.__output_maximum_delay, max(self.__inputs_maximum_delays)), __number_of_values):
            __X_matrix_row = list()
            __Y_matrix.append(output_values[k])
            for i in range(0, self.__output_maximum_delay):
                __X_matrix_row.append(-output_values[k-i-1])
            for j in range(self.__number_of_inputs):
                for i in range(self.__inputs_maximum_delays[j] - self.__input_minimum_delay + 1):
                    __X_matrix_row.append(list_of_input_values[j][k - i - self.__input_minimum_delay])
            if self.__offset:
                __X_matrix_row.append(1)
            __X_matrix.append(__X_matrix_row)

        __X_matrix = numpy.matrix(__X_matrix)
        __Y_matrix = numpy.matrix(__Y_matrix).transpose()

        parameters = (numpy.linalg.inv(__X_matrix.transpose() * __X_matrix) * __X_matrix.transpose() * __Y_matrix).squeeze().tolist()[0]
        self.__output_delay_parameters: dict[int, float] = dict()
        for i in range(self.__output_maximum_delay):
            self.__output_delay_parameters[i + 1] = -parameters[i]

        self.__inputs_delay_parameters: list[dict[int, float]] = [dict() for _ in range(self.__number_of_inputs)]
        parameters_index = self.__output_maximum_delay
        for i in range(self.__number_of_inputs):
            for j in range(0, self.__inputs_maximum_delays[i] - self.__input_minimum_delay + 1):
                self.__inputs_delay_parameters[i][j + self.__input_minimum_delay] = parameters[parameters_index]
                parameters_index += 1
        if self.__offset:
            self.__offset_value = parameters[-1]
        else:
            self.__offset_value = 0

    @property
    def input_delay_parameters(self):
        """Estimate parameter values multiplying to input variables.

        :return: list of dictionaries corresponding to each input with delay as key and parameter as value
        :rtype: list[dict[int, float]]
        """
        return self.__inputs_delay_parameters

    @input_delay_parameters.setter
    def input_delay_parameters(self, input_delay_parameters):
        """Set the parameter values multiplying to input variables.

        :param input_delay_parameters: list of dictionaries corresponding to each input with delay as key and parameter as value
        :type input_delay_parameters: list[dict[int, float]]
        """
        self.__input_delay_parameters = input_delay_parameters

    @property
    def output_delay_parameters(self):
        """Estimate parameter values multiplying to output variables.

        :return: dictionary corresponding  with delay as key and parameter as value.
        :rtype: dict[int, float]
        """
        return self.__output_delay_parameters

    @output_delay_parameters.setter
    def output_delay_parameters(self, output_delay_parameters):
        """Set the parameter values multiplying to output variables.

        :param output_delay_parameters: dictionary corresponding  with delay as key and parameter as value
        :type output_delay_parameters: list[float]
        """
        self.__output_delay_parameters = output_delay_parameters

    @property
    def offset_value(self) -> float:
        """Get the resulting offset value.

        :return: offset value. If offset has been set to false in the initializer, it will return 0.
        :rtype: float
        """
        return self.__offset_value

    @offset_value.setter
    def offset_value(self, offset_value):
        """Set an offset value.

        :param offset_value: offset value
        :type offset_value: float
        """
        self.__offset_value = offset_value

    @property
    def maximum_delay(self) -> int:
        """Estimate the the maximum delay considering delays in output but also in all inputs.

        :return: the maximum delay considering delays in output but also in all inputs
        :rtype: int
        """
        __maximum_delay = max(self.__output_delay_parameters)
        for input_delay_parameters in self.__inputs_delay_parameters:
            __maximum_delay = max(__maximum_delay, max(input_delay_parameters))
        return __maximum_delay

    def simulate(self, list_of_inputs_values, list_of_initial_output_values=None):
        """Simulate the ouput response using input values and estimated (with learn) linear regression.

        :param list_of_inputs_values: list of variable data to be used for input variables where size = number of input variables x number of data
        :type list_of_inputs_values: list[list[float]]
        :param list_of_initial_output_values: list of variable data to be used for output variables to initialize the linear regression, where size >= maximum delay (values corresponding to index > maximum delay are ignored). If an int is provided, the values will be iniatized with it and None, 0 will be used. Default to None
        :type list_of_initial_output_values: list[float]
        :return: simulated output with learnt linear regression
        :rtype: list[float]
        """
        number_of_values = len(list_of_inputs_values[0])
        if list_of_initial_output_values is None:
            estimated_output_values = [0 for i in range(self.maximum_delay+1)]
        elif type(list_of_initial_output_values) == float or type(list_of_initial_output_values) == int:
            estimated_output_values = [list_of_initial_output_values for i in range(self.maximum_delay + 1)]
        else:
            estimated_output_values = [list_of_initial_output_values[i] for i in range(self.maximum_delay+1)]
        self.__output_delay_impacts = {delay:0 for delay in self.__output_delay_parameters}
        self.__inputs_delay_impacts = [{delay:0 for delay in self.__inputs_delay_parameters[i]} for i in range(self.__number_of_inputs)]
        self.__offset_impact = 0
        self.__total_impact = 0
        for k in range(self.maximum_delay+1, number_of_values):
            estimated_output_value = 0
            for output_delay in self.__output_delay_parameters:
                term = self.__output_delay_parameters[output_delay] * estimated_output_values[k-output_delay]
                estimated_output_value += term
                self.__output_delay_impacts[output_delay] += abs(term)
                self.__total_impact += abs(term)
            for i in range(self.__number_of_inputs):
                for input_delay in self.__inputs_delay_parameters[i]:
                    term = self.__inputs_delay_parameters[i][input_delay] * list_of_inputs_values[i][k - input_delay]
                    estimated_output_value += term
                    self.__inputs_delay_impacts[i][input_delay] += abs(term)
                    self.__total_impact += abs(term)
            estimated_output_values.append(estimated_output_value + self.__offset_value)
            self.__offset_impact += abs(self.__offset_value)
            self.__total_impact += abs(self.__offset_value)
        return estimated_output_values

    def sliding(self, list_of_inputs_values, list_of_output_values, time_slice_size=24, minimum_time_slices=15, time_slice_memory: int = None, log: bool = True):
        """Simulate with sliding window jumping from time slice to time slice to learn new parameters of linear regression and predict output.

        :param list_of_inputs_values: list of variable data to be used for input variables where size = number of input variables x number of data
        :type list_of_inputs_values: list[list[float]]
        :param list_of_output_values: list of variable data to be used for output variables to initialize the linear regression, where size >= maximum delay (values corresponding to index > maximum delay are ignored). If an int is provided, the values will be iniatized with it and None, 0 will be used
        :type list_of_output_values: list[float]
        :param time_slice_size: size of the time slice (default: 24) usually corresponding to one day
        :type time_slice_size: int
        :param minimum_time_slices: the initial number of time slices used to learn parameters. If too small, it will generate a singular matrix error. Default to 15
        :type minimum_time_slices: int
        :param time_slice_memory: maximum number of time slices kept for learning parameters. If smaller then minimum_time_slices, time_slice_memory will be set to minimum_time_slices. Default is None, which means no memory limitation. Default to None
        :type time_slice_memory: int
        :param log: log results if True. Default is True
        :type log: bool
        :return: estimated output simulated per time slice
        :rtype: list[float]
        """
        if time_slice_memory is not None:
            time_slice_memory = max(time_slice_memory, minimum_time_slices)
        inputs_slices = [LinearRegression.__extract_inputs(k, (k + 1) * time_slice_size, list_of_inputs_values) for k in range(0, minimum_time_slices * time_slice_size, time_slice_size)]
        outputs_slices = [LinearRegression.__extract_outputs(k, (k + 1) * time_slice_size, list_of_output_values) for k in range(0, minimum_time_slices * time_slice_size, time_slice_size)]
        estimated_outputs = list_of_output_values[0:minimum_time_slices * time_slice_size].copy()
        for k in range(minimum_time_slices * time_slice_size, len(list_of_output_values), time_slice_size):
            merged_inputs = LinearRegression.__merge_inputs(inputs_slices)
            merged_outputs = LinearRegression.__merge_outputs(outputs_slices)
            self.learn(merged_inputs, merged_outputs)
            if log:
                print(self)
            inputs_slices.append(LinearRegression.__extract_inputs(k, k + time_slice_size, list_of_inputs_values))
            all_estimated_outputs = self.simulate(LinearRegression.__merge_inputs(inputs_slices), estimated_outputs)
            estimated_outputs.extend(all_estimated_outputs[-time_slice_size:])
            outputs_slices.append(LinearRegression.__extract_outputs(k, k + time_slice_size, list_of_output_values))
            if time_slice_memory is not None and len(outputs_slices) > time_slice_memory:
                inputs_slices = inputs_slices[-time_slice_memory:]
                outputs_slices = outputs_slices[-time_slice_memory:]
        return estimated_outputs[0:len(list_of_output_values)]

    @staticmethod
    def __extract_inputs(from_k, to_k, list_of_input_values):
        """Extract a slice of time for the input data.

        :param from_k: beginning of the time slice
        :type from_k: int
        :param to_k: end of the time slice
        :type to_k: int
        :param list_of_input_values: input data
        :type list_of_input_values: list[list[float]]
        :return: time slice of the input data
        :rtype: list[float]
        """
        extracted_inputs = list()
        for i in range(len(list_of_input_values)):
            extracted_inputs.append(list_of_input_values[i][from_k:to_k])
        return extracted_inputs

    @staticmethod
    def __extract_outputs(from_k: int, to_k: int, list_of_output_values):
        """Extract a slice of time for the outout data.

        :param from_k: beginning of the time slice
        :type from_k: int
        :param to_k: end of the time slice
        :type to_k: int
        :param list_of_input_values: output data
        :type list_of_output_values: list[float]
        :return: time slice of the output data
        :rtype: list[float]
        """
        return list_of_output_values[from_k:to_k]

    @staticmethod
    def __merge_inputs(list_of_inputs_slices):
        """Merge several input time slices into a single one.

        :param list_of_inputs_slices: time slices
        :type list_of_inputs_slices: list[list[list[float]]]
        :return: an unique time slice, which is the concatenation of the slices, respected to order with which they have been provided
        :rtype: list[list[float]]
        """
        merged_inputs = copy.deepcopy(list_of_inputs_slices[0])
        for j in range(1, len(list_of_inputs_slices)):
            for i in range(len(list_of_inputs_slices[0])):
                merged_inputs[i].extend(list_of_inputs_slices[j][i])
        return merged_inputs

    @staticmethod
    def __merge_outputs(list_of_output_slices):
        """Merge several output time slices into a single one.

        :param list_of_ioutput_slices: time slices
        :return: an unique time slice, which is the concatenation of the slices, respected to order with which they have been provided
        :rtype: list[float]
        """
        merged_outputs = copy.deepcopy(list_of_output_slices[0])
        for j in range(1, len(list_of_output_slices)):
            merged_outputs.extend(list_of_output_slices[j])
        return merged_outputs

    def __str__(self):
        """Return a descriptive string of the linear regression.

        :return: text representation
        """
        string = '%s{k} = ' % (self.__output_label)
        for delay in self.__output_delay_parameters:
            if self.__output_delay_parameters[delay] > 0:
                string += '+%f %s{k-%i} ' % (self.__output_delay_parameters[delay], self.__output_label, delay)
            else:
                string += '%f %s{k-%i} ' % (self.__output_delay_parameters[delay], self.__output_label, delay)
            if self.__total_impact != 0:
                string += '(%.2f%%)' % (100 * self.__output_delay_impacts[delay] / self.__total_impact)
        for i in range(self.__number_of_inputs):
            string += '\n\t\t\t...'
            for delay in self.__inputs_delay_parameters[i]:
                if self.__inputs_delay_parameters[i][delay] > 0:
                    if delay != 0:
                        string += '+%f %s{k-%i} ' % (self.__inputs_delay_parameters[i][delay], self.__input_labels[i], delay)
                    else:
                        string += '+%f %s{k} ' % (self.__inputs_delay_parameters[i][delay], self.__input_labels[i])
                else:
                    if delay != 0:
                        string += '%f %s{k-%i} ' % (self.__inputs_delay_parameters[i][delay], self.__input_labels[i], delay)
                    else:
                        string += '%f %s{k} ' % (self.__inputs_delay_parameters[i][delay], self.__input_labels[i])
                if self.__total_impact != 0:
                    string += '(%.2f%%)' % (100 * self.__inputs_delay_impacts[i][delay] / self.__total_impact)
        if self.__offset:
            if self.__offset_value > 0:
                string += '\n\t\t\t +%f' % self.__offset_value
            elif self.__offset_value < 0:
                string += '\n\t\t\t %f' % self.__offset_value
            if self.__offset_value != 0 and self.__total_impact != 0:
                string += '(%.2f%%)' % (100 * self.__offset_impact / self.__total_impact)
        return string

    def error_analysis(self, list_of_inputs_values, list_of_actual_output_values, list_of_estimated_output_values, maxlags=10, folder_name='log'):
        """Analyse the error of estimation by characterizing error, and analyzing correlations with input and auto-correlation of the output.

        :param list_of_inputs_values: input data values
        :type list_of_inputs_values: list[list[float]]
        :param list_of_actual_output_values: actual recorded output values
        :type list_of_actual_output_values: list[float]
        :param list_of_estimated_output_values: estimated values with the learnt linear regression
        :type list_of_estimated_output_values: list[float]
        :param maxlags: optional parameter used for cross correlation, default is 10.
        :type maxlags: int
        """
        number_of_parameters = len(self.__output_delay_parameters)
        for i in range(self.__number_of_inputs):
            number_of_parameters += len(self.__inputs_delay_parameters[i])
        if self.__offset:
            number_of_parameters += 1
        number_of_data = len(list_of_actual_output_values)
        output_errors = [list_of_actual_output_values[_] - list_of_estimated_output_values[_] for _ in range(number_of_data)]
        loss_function = (sum([error ** 2 for error in output_errors]) / len(output_errors)) / 2
        akaike_value = (1 + number_of_parameters / len(output_errors)) / (1 - number_of_parameters / len(output_errors)) * loss_function
        print('* Average output error = %f' % (sum(output_errors) / len(output_errors)))
        print('* Average absolute output error = %f' % (sum([abs(error) for error in output_errors]) / len(output_errors)))
        print('* LOSS function = %f' % loss_function)
        print('* AKAIKE value = %f' % akaike_value)
        print('* Max output error = %f' % max(output_errors))
        print('* Min output error = %f' % min(output_errors))
        print('* Standard deviation for output error = %f' % numpy.std(output_errors))
        sorted_output_errors = output_errors.copy()
        sorted_output_errors.sort()
        output_errors10 = sorted_output_errors[0: int(number_of_data / 10)]
        output_errors90 = sorted_output_errors[number_of_data - int(number_of_data/10): number_of_data]
        print('* 10%% lowest error average = %f' % (sum(output_errors10)/len(output_errors10)))
        print('* 90%% highest error average = %f' % (sum(output_errors90)/len(output_errors10)))

        fig = matplotlib.pyplot.figure()
        fig.suptitle('Output error Histogram')
        axes = fig.add_subplot(1, 1, 1)
        axes.hist(output_errors, bins=50)
        axes.set_ylabel('Frequency')
        fig.tight_layout()
        matplotlib.pyplot.savefig(folder_name+'/histogram.png')
        print('* Histogram')
        print('![histogram](histogram.png)')
        number_of_rows = math.ceil(math.sqrt(self.__number_of_inputs))
        number_of_columns = math.ceil(math.sqrt(self.__number_of_inputs))
        fig = matplotlib.pyplot.figure()
        fig.suptitle('trend analysis')
        for i in range(self.__number_of_inputs):
            axes1 = fig.add_subplot(number_of_rows, number_of_columns, i + 1)
            axes1.set_xlabel('input' + str(i))
            axes1.set_xlabel('time')
            axes1.set_ylabel('output errors', color='tab:red')
            axes1.plot([i for i in range(number_of_data)], output_errors, color='tab:red')
            axes1.tick_params(axis='y', labelcolor='tab:red')
            axes1.grid()
            axes2 = axes1.twinx()
            axes2.set_ylabel(self.__input_labels[i], color='tab:blue')
            axes2.plot([i for i in range(number_of_data)], list_of_inputs_values[i], color='tab:blue')
            axes2.tick_params(axis='y', labelcolor='tab:blue')
        fig.tight_layout()
        matplotlib.pyplot.savefig(folder_name+'/trends.png')
        print('* Trends')
        print('![trends](trends.png)')

        fig, axes = matplotlib.pyplot.subplots()  # auto-correlation
        fig.suptitle("Auto-correlation error")
        axes.acorr(output_errors, normed=True, usevlines=True, maxlags=maxlags)
        axes.set_xlim([-maxlags - 0.5, maxlags + 0.5])
        axes.grid()
        fig.tight_layout()
        matplotlib.pyplot.savefig(folder_name+'/autocorrelations.png')
        print('* Auto-correlations error')
        print('![correlations](autocorrelations.png)')

        fig, axes = matplotlib.pyplot.subplots()
        fig.suptitle('Cross correlation inputs-error analysis')
        for i in range(self.__number_of_inputs):
            axes = fig.add_subplot(number_of_rows, number_of_columns, i + 1)
            axes.set_xlabel(self.__input_labels[i])
            axes.xcorr(output_errors, list_of_inputs_values[i], normed=True, usevlines=True, maxlags=maxlags)
            axes.set_xlim([-maxlags - 0.5, 0.5])
            axes.grid()
        fig.tight_layout()
        matplotlib.pyplot.savefig(folder_name+'/cross-correlations.png')
        print('* Cross correlations inputs-errors')
        print('![correlations](cross-correlations.png)')

    def plot_zeros_poles(self, folder_name='log'):
        """Plot the zeros (roots of the numerator) and the poles (roots of the denominator) in a single figure with zeros as 'o' and poles as 'x'.

        The resulting figure is saved as 'log/zeros-poles.png' and a markdown string is returned on the standard output stream (with print).
        """
        print('## zeros-poles analysis')
        print('### poles')
        denominator = [0 for _ in range(self.maximum_delay + 1)]
        denominator[0] = 1
        for delay in self.__output_delay_parameters:
            denominator[delay] = - self.__output_delay_parameters[delay]
        poles = numpy.roots(denominator)
        zeros = list()
        for i in range(len(self.__input_labels)):
            print('### zeros for input %s' % self.__input_labels[i])
            numerator = [0 for j in range(self.maximum_delay + 1)]
            for delay in self.__inputs_delay_parameters[i]:
                numerator[delay] = self.__inputs_delay_parameters[i][delay]
            zeros.append(numpy.roots(numerator))
            print('* ', zeros[-1])

        number_of_plots = len(self.__input_labels)
        number_of_rows = math.floor(math.sqrt(number_of_plots))
        number_of_columns = math.ceil(number_of_plots / number_of_rows)
        fig = matplotlib.pyplot.figure()
        fig.suptitle('zeros-poles analysis')
        for i in range(0, len(self.__input_labels)):
            axes = fig.add_subplot(number_of_rows, number_of_columns, i + 1)
            axes.set_xlabel(self.__input_labels[i])
            # create the unit circle
            create_unit_circle = matplotlib.patches.Circle((0, 0), radius=1, fill=False, color='black', ls='dashed')
            axes.add_patch(create_unit_circle)
            # Plot the zeros and set marker properties
            plot_zeros = axes.plot(zeros[i].real, zeros[i].imag, 'go', ms=10)
            matplotlib.pyplot.setp(plot_zeros, markersize=10.0, markeredgewidth=1.0, markeredgecolor='k', markerfacecolor='g')
            # Plot the poles and set marker properties
            plot_poles = axes.plot(poles.real, poles.imag, 'rx', ms=10)
            matplotlib.pyplot.setp(plot_poles, markersize=12.0, markeredgewidth=3.0, markeredgecolor='r', markerfacecolor='r')
            axes.spines['left'].set_position('center')
            axes.spines['bottom'].set_position('center')
            axes.spines['right'].set_visible(False)
            axes.spines['top'].set_visible(False)
            # set the ticks
            radius = 1.5
            axes.axis('scaled')
            axes.axis([-radius, radius, -radius, radius])
            ticks = [-1, -.5, .5, 1]
            matplotlib.pyplot.xticks(ticks)
            matplotlib.pyplot.yticks(ticks)
            axes.grid()
        fig.tight_layout()
        matplotlib.pyplot.savefig(folder_name+'/zeros-poles.png')
        print('### Graphical representation')
        print('![zeros-poles](zeros-poles.png)')



"""This is a reader for openweathermap (https://openweathermap.org) historical weather files.

Author: stephane.ploix@grenoble-inp.fr
"""
from buildingenergy import timemg
import json

class OpenWeatherMapJsonReader:
    """Extract the content of a json openweather data file.

    :param json_filename: openweather data file in json format
    :param from_stringdate: initial date in format DD/MM/YYYY hh:mm:ss
    :param to_stringdate: final date in format DD/MM/YYYY hh:mm:ss
    :return: a tuple containing
        - city file_name
        - latitude in decimal north degree
        - longitude in decimal east degree
        - hourly time data variables as a dictionnary with variable file_name as a key
        - units as a dictionnary with variable file_name as a key
        - initial date as a string
        - final date as a string
    """

    def __init__(self, json_filename, from_stringdate=None, to_stringdate=None, sea_level_in_meter=290, albedo=.1, pollution=0.1):
        """Read data from an openweather map json file.

        :param json_filename: name of the openweathermap historical weather file
        :type json_filename: str
        :param from_stringdate: starting date for the data collection, defaults to None
        :type from_stringdate: str, optional
        :param to_stringdate: ending date for the data collection, defaults to None
        :type to_stringdate: str, optional
        :param sea_level_in_meter: sea level in meter of the site location, defaults to 290
        :type sea_level_in_meter: float, optional
        :param albedo: albedo at current site location (see https://en.wikipedia.org/wiki/Albedo), defaults to .1
        :type albedo: float, optional
        :param pollution: turbidity coefficient to model the air pollution at the current site location, defaults to 0.1
        :type pollution: float, optional
        """
        if from_stringdate is not None:
            from_epochtimems = timemg.stringdate_to_epochtimems(from_stringdate)
        if to_stringdate is not None:
            to_epochtimems = timemg.stringdate_to_epochtimems(to_stringdate)

        with open(json_filename) as json_file:
            weather_records = json.load(json_file)
            location = weather_records[0]['city_name']
            latitude_in_deg = float(weather_records[0]['lat'])
            longitude_in_deg = float(weather_records[0]['lon'])
            variable_names = ('epochtimems', 'temperature', 'wind_speed', 'wind_direction_in_deg', 'feels_like', 'humidity', 'pressure', 'cloudiness', 'temp_min', 'temp_max', 'description')
            variable_units = {'epochtimems': 'ms', 'temperature': 'celcius', 'wind_speed': 'm/s', 'wind_direction_in_deg': 'degree', 'feels_like': 'celcius', 'humidity': 'percent', 'pressure': 'hPa', 'cloudiness': 'percent', 'temp_min': 'celcius', 'temp_max': 'celcius', 'description': 'text'}
            self._site_weather_data = SiteWeatherData(location, latitude_in_deg, longitude_in_deg, variable_names, variable_units, sea_level_in_meter=sea_level_in_meter, albedo=albedo, pollution=pollution, _direct_call=False)

            for _record in weather_records:
                _epochtimems = int(_record['dt']) * 1000
                if from_stringdate is not None and _epochtimems >= from_epochtimems and from_stringdate is not None and _epochtimems <= to_epochtimems:
                    _temperature = float(_record['main']['temp'])
                    _wind_speed = float(_record['wind']['speed'])
                    _wind_direction_in_deg = float(_record['wind']['deg'])
                    _feels_like = float(_record['main']['feels_like'])
                    _humidity = float(_record['main']['humidity'])
                    _pressure = float(_record['main']['pressure'])
                    _cloudiness = float(_record['clouds']['all'])
                    _temp_min = float(_record['main']['temp_min'])
                    _temp_max = float(_record['main']['temp_max'])
                    _description = ''
                    for weather_description in _record['weather']:
                        _description += weather_description['description'] + ', '
                    self._site_weather_data._add_row(_epochtimems, _temperature, _wind_speed, _wind_direction_in_deg, _feels_like, _humidity, _pressure, _cloudiness, _temp_min, _temp_max, _description[:-2])

    @property
    def site_weather_data(self):
        """Return the site weather data.

        :return: a SideData object containing all the information of a site (see SiteWeatherData)
        :rtype: a buildingenergy.SiteWeatherData object.
        """
        return self._site_weather_data


class SiteWeatherData:
    """Gathers all the data related to a site dealing with location, albedo, pollution, timezone but also weather timedata coming from an openweather json file."""

    def __init__(self, location, latitude_in_deg, longitude_in_deg,
variable_names, variable_units, sea_level_in_meter=290, albedo=.1, timezone='Europe/Paris', pollution=0.1, _direct_call=True):
        """Create object containing data dealing with a specific site, including the weather data.

        :param location: name of the site
        :type location: str
        :param latitude_in_deg: latitude in East degree
        :type latitude_in_deg: float
        :param longitude_in_deg: longitude in North degree
        :type longitude_in_deg: float
        :param variable_names: name of the weather variables
        :type variable_names: tuple[str]
        :param variable_units: units of the weather variables
        :type variable_units: tuple[str]
        :param sea_level_in_meter: altitude of the site in meter from sea level, defaults to 290
        :type sea_level_in_meter: float, optional
        :param albedo: albedo of the site, defaults to .1
        :type albedo: float, optional
        :param timezone: timezone of the site, defaults to 'Europe/Paris'
        :type timezone: str, optional
        :param pollution: pollution coefficient between 0 and 1, defaults to 0.1
        :type pollution: float, optional
        :param _direct_call: internal use to prohibit direct calls of the initializer, defaults to True
        :type _direct_call: bool, optional
        :raises PermissionError: raised in case of direct use, the object must be created by OpenWeatherMapJsonReader
        """
        if _direct_call:
            raise PermissionError('SiteWeatherData cannot be called directly')
        self._location = location
        self._latitude_in_deg = latitude_in_deg
        self._longitude_in_deg = longitude_in_deg
        self._sea_level_in_meter = sea_level_in_meter
        self._albedo = albedo
        self._pollution = pollution
        self._timezone = timezone
        self._variable_names : tuple[str] = variable_names
        self._variable_units: tuple[str] = variable_units
        self._from_epochtimems = None
        self._to_epochtimems = None
        self._variable_data = dict()
        for variable_name in variable_names:
            self._variable_data[variable_name] = []
        self._cache = dict()
        self._cache_from_epochtimems = None
        self._cache_to_epochtimems = None

    @property
    def location(self):
        """Return the site name.

        :return: name of the site
        :rtype: str
        """
        return self._location

    @property
    def latitude_in_deg(self):
        """Return the latitude of the site in East degrees.

        :return: latitude of the site in East degrees
        :rtype: float
        """
        return self._latitude_in_deg

    @property
    def longitude_in_deg(self):
        """Return the longitude of the site in North degrees.

        :return: longitude of the site in North degrees
        :rtype: float
        """
        return self._longitude_in_deg

    @property
    def sea_level_in_meter(self):
        """Return the altitude of the site in meters from sea level.

        :return: altitude of the site
        :rtype: float
        """
        return self._sea_level_in_meter

    @property
    def albedo(self):
        """Return the albedo of the site.

        :return: albedo coefficient of the ground between 0 and 1
        :rtype: float
        """
        return self._albedo

    @property
    def pollution(self):
        """Return the air turbidity of the site as a coefficient.

        :return: pollution coefficient between 0 and 1
        :rtype: float
        """
        return self._pollution

    @property
    def timezone(self):
        """Return the site administrative time zone.

        :return: the timezone of the site
        :rtype: str
        """
        return self._timezone

    @property
    def variable_names(self):
        """Return the available weather variables.

        :return: list of the available weather variables
        :rtype: list[str]
        """
        return self._variable_names

    def unit(self, variable_name):
        """Return the unit of a variable.

        :param variable_name: file_name of the variable
        :type variable_name: str
        :return: unit of this variable
        :rtype: str
        """
        return self._variable_units[variable_name]

    @property
    def from_stringdate(self):
        """Return the starting data of the data collection.

        :return: first date where data are available in epoch time (ms)
        :rtype: int
        """
        return timemg.epochtimems_to_stringdate(self._from_epochtimems)

    @property
    def to_stringdate(self):
        """Return the ending date of the data collection.

        :return: last date where data are available
        :rtype: str
        """
        return timemg.epochtimems_to_stringdate(self._to_epochtimems)

    def _add_row(self, *variable_values):
        """Add a row to the weather data collection.

        Internal usage: must not be used directly

        :param variable_values: a list of values correspong to a row in the json openweather file
        :type variable_values: list[float]
        """
        epochtimems = variable_values[0]
        if self._from_epochtimems is None:
            self._from_epochtimems = epochtimems
            self._to_epochtimems = epochtimems
        elif epochtimems > self._to_epochtimems:
            self._to_epochtimems = epochtimems
        for i in range(len(variable_values)):
            self._variable_data[self._variable_names[i]].append(variable_values[i])

    def get(self, variable_name):
        """Return the data collection related to one variable.

        :param variable_name: variable file_name
        :type variable_name: str
        :return: list of float or str values corresponding to common dates for the specified variable
        :rtype: list[float or str]
        """
        if self._cache_from_epochtimems is not None and self._cache_from_epochtimems == self._from_epochtimems and self._cache_to_epochtimems == self._to_epochtimems and variable_name in self._cache:
            return self._cache[variable_name]
        _data = list()
        for i in range(len(self._variable_data['epochtimems'])):
            if self._from_epochtimems <= self._variable_data['epochtimems'][i] <= self._to_epochtimems:
                if variable_name == 'stringdate':
                    _data.append(timemg.epochtimems_to_stringdate(self._variable_data['epochtimems'][i]))
                elif variable_name == 'datetime':
                    _data.append(timemg.epochtimems_to_datetime(self._variable_data['epochtimems'][i]))
                else:
                    _data.append(self._variable_data[variable_name][i])
        if self._cache_from_epochtimems != self._from_epochtimems or self._cache_to_epochtimems != self._to_epochtimems:
            self._cache.clear()
        self._cache_from_epochtimems = self._from_epochtimems
        self._cache_to_epochtimems = self._to_epochtimems
        self._cache[variable_name] = _data
        return _data

    def day_degrees(self, temperature_reference=18, heat=True):
        """Compute heating or cooling day degrees and print in terminal the sum of day degrees per month.

        :param temperature_reference: reference temperature (default is 18°C)
        :param heat: True if heating, False if cooling
        :return: list of day dates as string, list of day average, min and max outdoor temperature and day degrees per day
        :rtype: [list[str], list[float], list[float], list[float], list[float]]
        """
        datetimes = self.get('datetime')
        stringdates = self.get('stringdate')
        temperatures = self.get('temperature')
        dd_months = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        month_names = ['January', 'February', 'March', 'April', 'May', 'June', 'July', 'August', 'September', 'October', 'November', 'December']
        day_stringdate_days = list()
        average_temperature_days = list()
        min_temperature_days = list()
        max_temperature_days = list()
        heat_degree_days = list()
        day_temperature = list()
        current_day = datetimes[0].day
        for k in range(len(datetimes)):
            if current_day == datetimes[k].day:
                day_temperature.append(temperatures[k])
            else:
                day_stringdate_days.append(stringdates[k-1].split(' ')[0])
                average_day_temperature = sum(day_temperature)/len(day_temperature)
                average_temperature_days.append(average_day_temperature)
                min_temperature_days.append(min(day_temperature))
                max_temperature_days.append(max(day_temperature))
                hdd = 0
                if heat:
                    if average_day_temperature < temperature_reference:
                        hdd = temperature_reference - average_day_temperature
                elif not heat:
                    if average_day_temperature > temperature_reference:
                        hdd = average_day_temperature - temperature_reference
                heat_degree_days.append(hdd)
                dd_months[datetimes[k].month-1] += hdd
                day_temperature = list()
            current_day = datetimes[k].day
        for i in range(len(dd_months)):
            print('day degrees', month_names[i], ': ', dd_months[i])
        return day_stringdate_days, average_temperature_days, min_temperature_days, max_temperature_days, heat_degree_days



"""A helper module dedicated to the deisgn of time-varying state space model approximated by bililear state space model.

Author: stephane.ploix@grenoble-inp.fr
"""

from abc import abstractmethod
import scipy.signal
import pickle
import os
import math
import numpy
from numpy import array, dot
from itertools import product


class StateMatrix:
    """State Matrix is representing a matrix with possible variations depending on influencing variable values.

    For instance, consider the M(x1, x2, x3,...) matrix where theta contains so-called influencing variables and xi in [0, 1]. To reduce calculations, M(x1, x2, x3,...) # M(0, 0, 0,...) + D{1, 0, 0,...} x1 + D{0, 1, 0,...} x2 +D{0, 0, 1,...} x3 + D{1, 1, 0,...} x1 x2 + D{1, 0, 1,...} x1 x3 + D{0, 1, 1,...} x2 x3 + D{1, 1, 1,...} x1 x2 x3 +...
    """

    def __init__(self, name, influencing_variable_names, matrix_at_maximum_influence_dict):
        """Create a state matrix with a nominal array. Variations according influencing variables are assumed to be additive. The nominal matrix is returned when all the variable values are set to 0. The maximum variation of the matrix wrt a given variable is obtained when the specified maximum value is reached.

        :param name: name of the time-varying matrix
        :type name: str
        :param influencing_variable_names: names of the normalized variables affecting the matrix
        :type influencing_variable_names: tuple[str]
        :param matrix_at_maximum_influence_dict: a dictionary of tuples with indices of influencing variables and resulting array when the influencing variables equal to 1 (maximum value)
        :type matrix_at_maximum_influence_dict: dict[tuple[int],numpy.array]
        """
        self.name = name
        self.influencing_variable_names = influencing_variable_names
        nominal_influence = tuple([0 for _ in range(len(influencing_variable_names))])
        self._nominal_array = numpy.copy(matrix_at_maximum_influence_dict[nominal_influence])
        self.delta_matrices_dict = dict()
        for influence in matrix_at_maximum_influence_dict.keys():
            if sum(influence) > 0:
                self.delta_matrices_dict[influence] = numpy.copy(matrix_at_maximum_influence_dict[influence]) - self._nominal_array

        delta_depth = 1
        while delta_depth <= len(influencing_variable_names):
            for influence in self.delta_matrices_dict:
                if sum(influence) < delta_depth:
                    for existing_influence in self.delta_matrices_dict:
                        if StateMatrix._is_included(existing_influence, influence, True):
                            self.delta_matrices_dict[influence] -= self.delta_matrices_dict[existing_influence]
            delta_depth += 1

    @staticmethod
    def _is_included(candidate_tuple, reference_tuple, accept_equality=True):
        """Check whether a tuple in included into another one.

        For instance, (0, 1, 1, 0) is included into (0, 1, 1, 1) but the opposite is false.
        If accept_equality is True (default) (0, 1, 1, 0) is included into (0, 1, 1, 0), if not accept_equality, it's no longer True.

        :param candidate_tuple: tuple to be tested
        :type candidate_tuple: tuple
        :param reference_tuple: reference tuple
        :type reference_tuple: tuple
        :param accept_equality: allows the equality of tuples, defaults to True
        :type accept_equality: bool, optional
        :return: result of the test
        :rtype: bool
        """
        for i in range(len(candidate_tuple)):
            if candidate_tuple[i] > reference_tuple[i]:
                return False
        if not accept_equality and sum(candidate_tuple) == sum(reference_tuple):
            return False
        return True

    def nominal(self):
        """Return the nominal state matrix.

        :return: nominal state matrix i.e. the one corresponding to nul influencing variables
        :rtype: numpy.array
        """
        return self._nominal_array

    def influencing_variable_names(self):
        """List of the variables influencing the matrix.

        :return: names of the influencing variables
        :rtype: tuple[str]
        """
        return self.influencing_variable_namesi

    def get(self, influencing_variable_values):
        """Return the matrix taking into account the influencing variable values.

        :param influencing_variable_values: variable names and their current value. If a variable is not present, its value is considered equal to 0. If a variable name has not been mentioned among the infuencing variables, it's ignored. defaults to None
        :type influencing_variable_values: list[float]
        :return: the current matrix value
        :rtype: numpy.array
        """
        current_influences = tuple([int(v > 0) for v in influencing_variable_values])
        matrix = numpy.copy(self.nominal())
        for influences in self.delta_matrices_dict:
            if StateMatrix._is_included(influences, current_influences):
                coef = 1
                for i in range(len(influences)):
                    if influences[i]:
                        coef = coef * influencing_variable_values[i]
                matrix += coef*self.delta_matrices_dict[influences]
        return matrix

    def __str__(self):
        """Return a description of the state matrix.

        :return: description of the state matrix
        :rtype: str
        """
        string = '\nmatrix %s =' % self.name
        for influences in self.delta_matrices_dict:
            variables = ''
            for i in range(len(influences)):
                if influences[i]:
                    variables += self.influencing_variable_names[i] + '*'
            string += '\n' + variables + '\n' + str(self.delta_matrices_dict[influences])
        return string

class Model:
    """Describe a linear time-varying model based on state space representation."""

    def __init__(self, sampling_time_in_secs, input_variable_names, influencing_variable_names, possible_action_variable_names, output_variable_names, power_gain_variable_name, display):
        """Create a linear time-varying dynamic model described by one or several alternative state space representation sharing the same linear inputs, the same inluencing variables affacting the state matrices and the same output. Parameter can be modified within bounds and a kind of filter is also recorded to transform measurement csv file and weather file into data.

        :param sampling_time_in_secs: the sampling time in seconds for the model to operate
        :type sampling_time_in_secs: int
        :param input_variable_names: list of the input variables intervening in a linear way
        :type input_variable_names: list[str]
        :param influencing_variable_names: list of the input variables influencing the state matrices i.e. corresponding to nonlinearities
        :type influencing_variable_names: list[str]
        :param influencing_variable_max_values: maximum values, respecting the order of the listed influencing variables. It corresponds to the maximum matrix deviations
        :typr influencing_variable_max_values: list[float]
        :param state_variable_names: list of the state space variables
        :type state_variable_names: list[str]
        :param output_variable_names: list of the output variables
        :type output_variable_names: list[str]
        :param power_gain_variable_name: name of the variable standing for the power gain. It must be part of the input variables
        :type power_gain_variable_name: str
        """
        self.sampling_time_in_secs = sampling_time_in_secs
        self._parameters_dict = dict()
        self.input_variable_names = input_variable_names
        self.influencing_variable_names = influencing_variable_names
        self.possible_action_variable_names = possible_action_variable_names
        self.output_variable_names = output_variable_names
        self.power_gain_index = input_variable_names.index(power_gain_variable_name)
        self.representation_state_variables = dict()
        self.state_variable_names: list[str] = None
        self.state_matrice_generator: function = None
        self.parameter_calculations(display)

    @abstractmethod
    def register_data(self, data_container, openweather_file_name):
        """Store measurement data and weather data read respectively from a DataContainer and an OpenWeather objects.

        The DataContainer should be saved as self.data_container to get the plotting capabilities. The useful data must be stored into a [str, list[float]] dictionnary: self._data. Weather data from the provided file name will be added to the data container if dates match measurements data

        :param data_container: data container containing measurement and weather data to work with the model
        :type data_container: measurements.DataContainer
        :param openweather_file_name: name of the open weather json file containing weather data
        :type openweather_file_name: str
        """
        raise NotImplemented

    def data(self, measurement_name):
        """Return measurements corresponding to a name.

        :param measurement_name: name of the measurement
        :type measurement_name: str
        :return: measurement data
        :rtype: list[float]
        """
        return self._data[measurement_name]

    @abstractmethod
    def parameter_calculations(self, display):
        """Abstract method that must be implemented to calculate model parameters and possibly record them with self.param and self.set_bound for parameter adjustement."""
        raise NotImplemented

    def register_state_space_representation(self, state_variables, state_matrice_generator):
        """Inherited method that register state space representations matching the specified input, influence and output variables.

        :param state_variables: list of the state variable names
        :type: tuple[str]
        :param matrice_generator: a function generating 4 continuous time state matrices A, B, C and D, each one represented by a StateMatrix
        :type matrice_generator: function
        """
        self.state_variable_names = state_variables
        self.state_matrice_generator = state_matrice_generator

    def initialize(self, use_state_observer=False):
        """Update the calculation of state matrices.

        Useful if parameters have been updated. It calls the init method of the subclass for specific calculation that returns an initial state vector.

        :param use_state_observer: True if a state observer should be used, default to False.
        :type use_state_observer: bool
        """
        self.Ad_state_matrix, self.Bd_state_matrix, self.Cd_state_matrix, self.Dd_state_matrix = self.discrete_state_matrices()
        if use_state_observer:
            self.K = self.matrix_K_state_observer()
        return self.init()

    @abstractmethod
    def matrix_K_state_observer(self):
        """Return the gain matrix K for the state observer. (abstract method)

        :return: the gain matrix
        :rtype: numpy.array"""
        raise NotImplemented

    @abstractmethod
    def init(self):
        """Use to initialize a simulation (intial state vector,...).

        :return: initial state vactor
        :rtype: numpy.array
        """
        pass

    @abstractmethod
    def register_data(self, measurement_file_name, weather_file_name, skiprows=0, nrows=None):
        """Store measurement data and weather data read respectively from a DataContainer and an OpenWeather objects. The DataContainer should be saved as self.data_container to get the plotting capabilities. The useful data must be stored into a [str, list[float]] dictionnary: self._data.

        :param measurement_file_name: csv file name containing measurement data from office
        :type measurement_file_name: str
        :param weather_file_name: json file name coming from openweather
        :type weather_file_name: str
        :param skiprows: list-like or integer Row numbers to skip (0-indexed) or number of rows to skip (int) at the start of the file
        :type skiprows: int, default is 0
        :param nrows: Number of rows of file to read. Useful for reading pieces of large files. Default is None
        :type nrows: int
        """
        raise NotImplemented

    def save_parameters(self, pickle_file_name):
        """Save current parameter values into a pickle file.

        :param pickle_file_name: name of the pickle file
        :type pickle_file_name: str
        """
        name_value_bounds_dict = dict()
        for name in self._parameters_dict:
            name_value_bounds_dict[name] = [self._parameters_dict[name].val, self._parameters_dict[name].bounds]
        with open(pickle_file_name, "wb") as file:
            pickle.dump(name_value_bounds_dict, file)

    def load_parameters(self, parameters_filename):
        """Load parameter values from a pickle file: these values will replace the nominal ones.

        :param parameters_filename: name of the file, defaults to 'best_parameters.p'
        :type parameters_filename: str, optional
        """
        if os.path.exists(parameters_filename):
            self._parameters_dict.clear()
            with open(parameters_filename, 'rb') as file:
                name_value_bounds_dict = pickle.load(file)
            for name in name_value_bounds_dict:
                self.param(name, name_value_bounds_dict[name][0], name_value_bounds_dict[name][1])

    class _Parameter:
        """Inner class containing data about a parameter."""

        def __init__(self, _name_parameter_dict, name, value=None, bounds=None):
            """Create a parameter and register it into the dictionary of all registered parameters.

            :param _name_parameter_dict: dictionary of existing parameters
            :type _name_parameter_dict: dict[str, buildingenergy.Model._Parameter]
            :param name: name of the parameter
            :type name: str
            :param value: current value for the parameter, defaults to None
            :type value: float, optional
            :param bounds: extrement possible values for the parameter, defaults to None
            :type bounds: tuple[float], optional
            """
            self._name_parameter_dict = _name_parameter_dict
            self._name = name
            self._value = value
            self._initial_value = value
            self._bounds = bounds
            self._name_parameter_dict[name] = self

        @property
        def name(self):
            """Return the name of the parameter.

            :return: name of the parameter
            :rtype: str
            """
            return self._name

        def __call__(self, inf, sup):
            self._bounds = (inf, sup)

        @property
        def val(self):
            """Return the current parameter value.

            :return: the current parameter value
            :rtype: float
            """
            return self._value

        @val.setter
        def val(self, value):
            """Set parameter value.

            :param value: value
            :type value: float
            """
            self._value = value

        @property
        def bounds(self):
            """Return bounds for possible parameter values.

            :return: tuple with inferior and superior bounds
            :rtype: tuple[float]
            """
            return self._bounds

        @bounds.setter
        def bounds(self, bounds):
            """Set bounds for possible parameter values.

            :param bounds: [description]
            :type bounds: [type]
            """
            self._bounds = bounds

        @property
        def all(self):
            """Return the names of the parameters.

            :return: list of names
            :rtype: list[st]
            """
            return self._name_parameter_dict

        @all.setter
        def all(self, parameters_dict):
            """Set all the parameters at once.

            :param parameters_dict: dictionary with parameter names as keys and Parameter object as values
            :type parameters_dict: dict[str, Parameter]
            """
            self._name_parameter_dict = parameters_dict

        def __str__(self):
            """Return a description of a parameter.

            :return: a descriptive string.
            :rtype: str
            """
            if self.bounds:
                if self._initial_value is None:
                    return '%s=%f in (%f,%f)' % (self._name, self._value, self._bounds[0], self._bounds[1])
                else:
                    return '%s=%f (init. %f) in (%f,%f)' % (self._name, self._value, self._initial_value, self._bounds[0], self._bounds[1])
            else:
                if self._initial_value is None:
                    return '%s=%f' % (self._name, self._value)
                else:
                    return '%s=%f (%f)' % (self._name, self._value, self._initial_value)

    def param(self, parameter_name, value=None, bounds=None):
        """Create a constant potentially ajustable parameter whose value can be invariant or can vary within bounds.

        :param parameter_name: name of the parameter
        :type parameter_name: str
        :param value: nominal value of the parameter
        :type value: float
        :param bounds: interval for the possible parameter values. If None, it's value will not be adjusted. Defaults to None.
        :type bounds: tuple[float], optional
        """
        return Model._Parameter(self._parameters_dict, parameter_name, value, bounds)

    def pval(self, parameter_names):
        """Return requested parameter values.

        :param parameter_names: name of the requested parameter values
        :type parameter_names: str or list[str]
        :return: the requested parameter values as a float or a list of float, depending on how the parameter names have been specified
        :rtype: float or list[float]
        """
        if type(parameter_names) == str:
            return self._parameters_dict[parameter_names].val
        else:
            pvals = list()
            for parameter_name in parameter_names:
                pvals.append(self._parameters_dict[parameter_name].val)
            return pvals

    def pbound(self, parameter_names):
        """Return requested parameter bounds.

        :param parameter_names: name(s) of the requested parameter bounds
        :type parameter_names: str or list[str]
        :return: parameter bounds as tuples or None value if the parameter cannot be adjusted
        :rtype: tuple[float] or list[tuple[float]]
        """
        if type(parameter_names) == str:
            return self._parameters_dict[parameter_names].bounds
        else:
            _pbounds = list()
            for parameter_name in parameter_names:
                _pbounds.append(self._parameters_dict[parameter_name].bounds)
            return _pbounds

    @property
    def parameters(self):
        """Return the list of recorded parameter names.

        :return: list of recorded parameter names
        :rtype: list[str]
        """
        return list(self._parameters_dict.keys())

    @property
    def adjustables(self):
        """List of adjustable parameter names i.e. the one whose bounds are not None.

        :return: list of adjustable parameter names
        :rtype: list[str]
        """
        _adjustable_parameters = list()
        for parameter in self._parameters_dict:
            if self._parameters_dict[parameter].bounds is not None:
                _adjustable_parameters.append(parameter)
        return _adjustable_parameters

    @adjustables.setter
    def adjustables(self, values):
        """Set all the values of adjustable parameters at once.

        :param values: list of values ordered according to the list of names returned by the methods 'self.parameters'.
        :type values: list[float]
        """
        _adjustables = self.adjustables
        for i in range(len(_adjustables)):
            self._parameters_dict[_adjustables[i]].val = values[i]

    def discrete_state_matrices(self):
        """Discretized registered state matrices (with method self.register_state_space_representation()).

        It's a function whose prototype must comply with compute_matrices(model: Model, influencing_variable_values: list[float]). It receives a model where registered parameters recorded by param can be obtained by model.pval('parameter_name'). The influencing variable values are provided according to the order given at the initialisation of the Model object. It returns the state space matrices obtained for the specified values of influencing variables: it must return the time continuous matrices A, B, C, D as a list[list[float]]].

        :return: matrices of the discrete time state space model taking into account linearized influence of influencing variables
        :rtype: tuple[StateMatrix]
        """
        list_of_influences =  list(product(range(0, 2), repeat=len(self.influencing_variable_names)))
        influenced_Ad_dict = dict()
        influenced_Bd_dict = dict()
        influenced_Cd_dict = dict()
        influenced_Dd_dict = dict()
        for influences in list_of_influences:
            A_tab, B_tab, C_tab, D_tab = self.state_matrice_generator(influences)
            Ad, Bd, Cd, Dd, _ = scipy.signal.cont2discrete((array(A_tab), array(B_tab), array(C_tab), array(D_tab)), self.sampling_time_in_secs, method='zoh')
            influenced_Ad_dict[influences] = Ad
            influenced_Bd_dict[influences] = Bd
            influenced_Cd_dict[influences] = Cd
            influenced_Dd_dict[influences] = Dd
        Ad_state_matrix = StateMatrix('Ad', self.influencing_variable_names, influenced_Ad_dict)
        Bd_state_matrix = StateMatrix('Bd', self.influencing_variable_names, influenced_Bd_dict)
        Cd_state_matrix = StateMatrix('Cd', self.influencing_variable_names, influenced_Cd_dict)
        Dd_state_matrix = StateMatrix('Dd', self.influencing_variable_names, influenced_Dd_dict)

        return Ad_state_matrix, Bd_state_matrix, Cd_state_matrix, Dd_state_matrix

    @abstractmethod
    def computeU(self, k: int, state_vector, influencing_variable_values, actions={}):
        """Compute the vector U of the state space representation together with the heating power, that can be the one that has been recorded or deduced from the temperature setpoint (assuming a perfect controller, with a one hour perspective).

        :param k: hour index
        :type k: int
        :param state_vector: current state vector used to determine the heating power in case a setpoint is specified in the actions
        :type state_vector: numpy.array
        :param influencing_variable_values: list of values for influencing variable ordered according to the order given at initialization
        :type influencing_variable_values: list[float]
        :param actions: list of current actions performaed on the system. It might contains in particular here 'heating_power' in case it's directly controlled, and 'temperature_setpoint' in case of an indirect control of the heating power by specifying the temperature setpoint
        :type: dict[str, float]
        :return: the vector U of the state space representation together with the heating power and an array with current output values
        :rtype: tuple[array, float]
        """
        raise NotImplemented

    def stepX(self, k: int, state_vector, observer_state_vector, influencing_variable_values, input_variables):
        """Compute the next state variables according to the U vector and the specified actions.

        :param k: step number
        :type k: int
        :param state_vector: the current state vector usually denoted by X
        :type state_vector: numpy.array
        :param observer_state_vector: the current state vector of state observer, None if there is no state observer.
        :type observer_state_vector: numpy.array
        :param influencing_variable_values: list of the current influencing variable values ordered according to the list provided at Model creation.
        :type influencing_variable_values: list[float]
        :return: the next step state vector X (X_{k+1})
        :param input_variables: values of the U vector in state space model.
        :rtype: numpy.array
        """
        state_vector = dot(self.Ad_state_matrix.get(influencing_variable_values), state_vector) + dot(self.Bd_state_matrix.get(influencing_variable_values), input_variables)
        if observer_state_vector is not None:
            Ymeasured = array([[self.data(name)[k]] for name in self.output_variable_names])
            observer_state_vector = dot(self.Ad_state_matrix.get(influencing_variable_values) - dot(self.K, self.Cd_state_matrix.get(influencing_variable_values)), observer_state_vector) + dot(self.Bd_state_matrix.get(influencing_variable_values) - dot(self.K, self.Dd_state_matrix.get(influencing_variable_values)), input_variables) + dot(self.K, Ymeasured)
        return state_vector, observer_state_vector

    def computeY(self, k, state_vector, U, influencing_variables):
        """Compute the output vector Y of the state space representation.

        :param k: step number
        :type k: int
        :param state_vector: the current state vector usually named X
        :type state_vector: numpy.array
        :param U: the vector U of the state space representation
        :type U: numpy.array
        :param influencing_variables: list of the current influencing variables affecting the state matrices.
        :type influencing_variables: list[float]
        :return: the output vector Y
        :rtype: numpy.array
        """
        if influencing_variables is None:
            influencing_variables = tuple([0 for _ in range(len(self.influencing_variable_names))])
        return dot(self.Cd_state_matrix.get(influencing_variables), state_vector) + dot(self.Dd_state_matrix.get(influencing_variables), U)

    def __str__(self):
        """Return a description of the model.

        :return: A desciption as a string.
        :rtype: str
        """

        string = '* input variables: '
        for input_variable_name in self.input_variable_names:
            string += '%s,' % input_variable_name
        string += '\n* influencing variables: '
        for influencing_variable_name in self.influencing_variable_names:
            string += '%s,' % influencing_variable_name
        if self.state_variable_names is None:
            string
        string += '\n* state variables: '
        if self.state_variable_names is None:
            string += 'Not defined'
        else:
            for state_variable_name in self.state_variable_names:
                string += '%s,' % state_variable_name
        string += '\n* output variables: '
        for output_variable_name in self.output_variable_names:
            string += '%s,' % output_variable_name
        string += '\n* parameters: '
        for parameter_name in self._parameters_dict:
            string += '%s=%f' % (parameter_name, self._parameters_dict[parameter_name].val)
            if self._parameters_dict[parameter_name].bounds is not None:
                string += '(%f,%f)' % (self._parameters_dict[parameter_name].bounds[0], self._parameters_dict[parameter_name].bounds[1])
            string += "\n"
        return string


class Preference:
    """Provide de a model of the occupants'preferences. It deals with thermal comfort, air quality, number of home configuration changes, energy cost, icone,..."""

    def __init__(self, preferred_temperatures=(21, 23), extreme_temperatures=(18, 26), preferred_CO2_concentration=(500, 1500), temperature_weight_wrt_CO2: float = 0.5, power_weight_wrt_comfort: float = 0.5e-3):
        """Definition of comfort regarding  number of required actions by accupants, temperature and CO2 concentration, but also weights between cost and comfort, and between thermal and air quality comfort.

        :param preferred_temperatures: preferred temperature range, defaults to (21, 23)
        :type preferred_temperatures: tuple, optional
        :param extreme_temperatures: limits of acceptable temperatures, defaults to (18, 26)
        :type extreme_temperatures: tuple, optional
        :param preferred_CO2_concentration: preferred CO2 concentration range, defaults to (500, 1500)
        :type preferred_CO2_concentration: tuple, optional
        :param temperature_weight_wrt_CO2: relative importance of thermal comfort wrt air quality (1 means only temperature is considered), defaults to 0.5
        :type temperature_weight_wrt_CO2: float, optional
        :param power_weight_wrt_comfort: relative importance of energy cost wrt comfort (1 means only energy cost is considered), defaults to 0.5e-3
        :type power_weight_wrt_comfort: float, optional
        """
        self.preferred_temperatures = preferred_temperatures
        self.extreme_temperatures = extreme_temperatures
        self.preferred_CO2_concentration = preferred_CO2_concentration
        self.temperature_weight_wrt_CO2 = temperature_weight_wrt_CO2
        self.power_weight_wrt_comfort = power_weight_wrt_comfort

    def change_dissatisfaction(self, occupancy, action_set):
        """Compute the ratio of the number of hours where occupants have to change their home configuration divided by the number of hours with presence.

        :param occupancy: a vector of occupancies
        :type occupancy: list[float]
        :param action_set: different vectors of actions
        :type action_set: tuple[list[float]]
        :return: the number of hours where occupants have to change their home configuration divided by the number of hours with presence
        :rtype: float
        """
        number_of_changes = 0
        number_of_presences = 0
        previous_actions = [actions[0] for actions in action_set]
        for k in range(len(occupancy)):
            if occupancy[k] > 0:
                number_of_presences += 1
                for i in range(len(action_set)):
                    actions = action_set[i]
                    if actions[k] != previous_actions[i]:
                        number_of_changes += 1
                        previous_actions[i] = actions[k]
        return number_of_changes / number_of_presences if number_of_presences > 0 else 0

    def thermal_comfort_dissatisfaction(self, temperatures, occupancies):
        """Compute average dissatisfaction regarding thermal comfort: 0 means perfect and greater than 1 means not acceptable. Note that thermal comfort is only taken into account if occupancy > 0, i.e. in case of presence.

        :param temperatures: vector of temperatures
        :type temperatures: list[float]
        :param occupancies: vector of occupancies (number of people per time slot)
        :type occupancies: list[float]
        :return: average dissatisfaction regarding thermal comfort
        :rtype: float
        """
        if type(temperatures) is not list:
            temperatures = [temperatures]
            occupancies = [occupancies]
        dissatisfaction = 0
        for i in range(len(temperatures)):
            if occupancies[i] != 0:
                if temperatures[i] < self.preferred_temperatures[0]:
                    dissatisfaction += (self.preferred_temperatures[0] - temperatures[i]) / (self.preferred_temperatures[0] - self.extreme_temperatures[0])
                elif temperatures[i] > self.preferred_temperatures[1]:
                    dissatisfaction += (temperatures[i] - self.preferred_temperatures[1]) / (self.extreme_temperatures[1] - self.preferred_temperatures[1])
        return dissatisfaction / len(temperatures)

    def air_quality_dissatisfaction(self, CO2_concentrations, occupancies):
        """Compute average dissatisfaction regarding air quality comfort: 0 means perfect and greater than 1 means not acceptable. Note that air quality comfort is only taken into account if occupancy > 0, i.e. in case of presence.

        :param CO2_concentrations: vector of CO2 concentrations
        :type CO2_concentrations: list[float]
        :param occupancies: vector of occupancies (number of people per time slot)
        :type occupancies: list[float]
        :return: average dissatisfaction regarding air quality comfort
        :rtype: float
        """
        if type(CO2_concentrations) is not list:
            CO2_concentrations = [CO2_concentrations]
            occupancies = [occupancies]
        dissatisfaction = 0
        for i in range(len(CO2_concentrations)):
            if occupancies[i] != 0:
                dissatisfaction += max(0., (CO2_concentrations[i] - self.preferred_CO2_concentration[0]) / (self.preferred_CO2_concentration[1] - self.preferred_CO2_concentration[0]))
        return dissatisfaction / len(CO2_concentrations)

    def comfort_dissatisfaction(self, temperatures, CO2_concentrations, occupancies):
        """Compute the comfort weighted dissatisfaction that combines thermal and air quality dissatisfactions: it uses the thermal_dissatisfaction and air_quality_dissatisfaction methods.

        :param temperatures: list of temperatures
        :type temperatures: list[float]
        :param CO2_concentrations: list of CO2 concentrations
        :type CO2_concentrations: list[float]
        :param occupancies: list of occupancies
        :type occupancies: list[float]
        :return: the global comfort dissatisfaction
        :rtype: float
        """
        return self.temperature_weight_wrt_CO2 * self.thermal_comfort_dissatisfaction(temperatures, occupancies) + (1 - self.temperature_weight_wrt_CO2) * self.air_quality_dissatisfaction(CO2_concentrations, occupancies)

    def cost(self, Pheat, kWh_price=.13):
        """Compute the heating cost.

        :param Pheat: list of heating power consumptions
        :type Pheat: list[float]
        :param kWh_price: tariff per kWh, defaults to .13
        :type kWh_price: float, optional
        :return: energy cost
        :rtype: float
        """
        if type(Pheat) is not list:
            Pheat = [Pheat]
        return sum(Pheat) / 1000 * kWh_price

    def icone(self, CO2_concentration, occupancy):
        """Compute the ICONE indicator dealing with confinement regarding air quality.

        :param CO2_concentration: list of CO2 conncentrations
        :type CO2_concentration: list[float]
        :param occupancy: list of occupancies
        :type occupancy: list[float]
        :return: value between 0 and 5
        :rtype: float
        """
        n_presence = 0
        n1_medium_containment = 0
        n2_high_containment = 0
        for k in range(len(occupancy)):
            if occupancy[k] > 0:
                n_presence += 1
                if 1000 <= CO2_concentration[k] < 1700:
                    n1_medium_containment += 1
                elif CO2_concentration[k] >= 1700:
                    n2_high_containment += 1
        f1 = n1_medium_containment / n_presence if n_presence > 0 else 0
        f2 = n2_high_containment / n_presence if n_presence > 0 else 0
        return 8.3 * math.log10(1 + f1 + 3 * f2)

    def assess(self, Pheater, temperatures, CO2_concentrations, occupancies) -> float:
        """Compute the global objective to minimise including both comforts and energy cost for heating.

        :param Pheater: list of heating powers
        :type Pheater: list[float]
        :param temperatures: list of temperatures
        :type temperatures: list[float]
        :param CO2_concentrations: list of CO2 concentrations
        :type CO2_concentrations: list[float]
        :param occupancies: list of occupancies
        :type occupancies: list[float]
        :return: objective value
        :rtype: float
        """
        return self.cost(Pheater) * self.power_weight_wrt_comfort + (1 - self.power_weight_wrt_comfort) * self.comfort_dissatisfaction(temperatures, CO2_concentrations, occupancies)

    def print_assessment(self, Pheat, temperatures, CO2_concentrations, occupancies, action_sets):
        """Print different indicateurs to appreciate the impact of a series of actions.

        :param Pheat: list of heating powers
        :type Pheat: list[float]
        :param temperatures: list of temperatures
        :type temperatures: list[float]
        :param CO2_concentrations: list of CO2 concentrations
        :type CO2_concentrations: list[float]
        :param occupancies: list of occupancies
        :type occupancies: list[float]
        :param actions: list of actions
        :type actions: tuple[list[float]]
        """
        print('- global objective: %s' % self.assess(Pheat, temperatures, CO2_concentrations, occupancies))
        print('- average thermal dissatisfaction: %.2f%%' % (self.thermal_comfort_dissatisfaction(temperatures, occupancies) * 100))
        print('- average CO2 dissatisfaction: %.2f%%' % (self.air_quality_dissatisfaction(CO2_concentrations, occupancies) * 100))
        print('- ICONE: %.2f' % (self.icone(CO2_concentrations, occupancies)))
        print('- average comfort dissatisfaction: %.2f%%' % (self.comfort_dissatisfaction(temperatures, CO2_concentrations, occupancies) * 100))
        print('- change dissatisfaction (number of changes / number of time slots with presence): %.2f%%' % (self.change_dissatisfaction(occupancies, action_sets) * 100))
        print('- heating cost: %s' % self.cost(Pheat))

        temperatures_when_presence = list()
        CO2_concentrations_when_presence = list()
        for i in range(len(occupancies)):
            if occupancies[i] > 0:
                temperatures_when_presence.append(temperatures[i])
                CO2_concentrations_when_presence.append(CO2_concentrations[i])
        if len(temperatures_when_presence) > 0:
            temperatures_when_presence.sort()
            CO2_concentrations_when_presence.sort()
            office_temperatures_estimated_presence_lowest = temperatures_when_presence[:math.ceil(len(temperatures_when_presence) * 0.1)]
            office_temperatures_estimated_presence_highest = temperatures_when_presence[math.floor(len(temperatures_when_presence) * 0.9):]
            office_co2_concentrations_estimated_presence_lowest = CO2_concentrations_when_presence[:math.ceil(len(CO2_concentrations_when_presence) * 0.1)]
            office_co2_concentrations_estimated_presence_highest = CO2_concentrations_when_presence[math.floor(len(CO2_concentrations_when_presence) * 0.9):]
            print('- average temperature during presence:', sum(temperatures_when_presence) / len(temperatures_when_presence))
            print('- average 10% lowest temperature during presence:', sum(office_temperatures_estimated_presence_lowest) / len(office_temperatures_estimated_presence_lowest))
            print('- average 10% highest temperature during presence:', sum(office_temperatures_estimated_presence_highest) / len(office_temperatures_estimated_presence_highest))
            print('- average CO2 concentration during presence:', sum(CO2_concentrations_when_presence) / len(CO2_concentrations_when_presence))
            print('- average 10% lowest CO2 concentration during presence:', sum(office_co2_concentrations_estimated_presence_lowest) / len(office_co2_concentrations_estimated_presence_lowest))
            print('- average 10% highest CO2 concentration during presence:', sum(office_co2_concentrations_estimated_presence_highest) / len(office_co2_concentrations_estimated_presence_highest))

    def __str__(self):
        """Return a description of the defined preferences.

        :return: a descriptive string of characters.
        :rtype: str
        """
        string = 'preference: temperature in %f<%f-%f>%f, concentrationCO2 %f>%f\n' % (self.extreme_temperatures[0], self.preferred_temperatures[0], self.preferred_temperatures[1], self.extreme_temperatures[1], self.preferred_CO2_concentration[0], self.preferred_CO2_concentration[1])
        string += '%.3f * cost + %.3f disT + %.3f disCO2' % (self.power_weight_wrt_comfort, (1-self.power_weight_wrt_comfort) * self.temperature_weight_wrt_CO2, (1-self.power_weight_wrt_comfort) * (1-self.temperature_weight_wrt_CO2))
        return string


"""Dynamic programming for the generation of optimal energy management strategies."""

import buildingenergy.model
import numpy


class Possibilities:
    """Represent a node in digraph representing a set of possible actions at a given timestep."""

    def __init__(self, value_lists, ancestor=None, value: float=None):
        """Initialize a node.

        :param value_lists: tuples of possible values respectively related to the possible controls
        :type value_lists: list[tuple[float]]
        :param ancestor: a node of possibilities at the previous sample time, which led to this node, defaults to None
        :type ancestor: Possibilities, optional
        :param value: values in term of optimization objective to be minimized, defaults to None
        :type value: float, optional
        :raises ValueError: raise an error if one of the 3 tuples of possible values is empty i.e. no one action is possible
        """
        if ancestor is None:
            Possibilities.root_node = self
            Possibilities.terminal_nodes = list()
            for i in range(len(value_lists)):
                if len(value_lists[i]) == 0:
                    raise ValueError('Generation of possibilities is impossible because of the empty set in position %i', i)
            self._rank = len(value_lists)
            Possibilities.full_rank = len(value_lists)
        else:
            self._rank = ancestor._rank - 1
            self._value =  value
        self._ancestor = ancestor
        self._descendants = list()
        self._value_lists = value_lists
        if len(self._value_lists) > 0:
            for value in value_lists[0]:
                self._descendants.append(Possibilities(value_lists[1:], self, value))
        else:
            Possibilities.terminal_nodes.append(self)

    def _val(self):
        """Return the sum of the objective values of the Possibilities bloodline leading to this node.

        :return: the objective value of the bloodline leading to this node
        :rtype: float
        """
        ancestor = self
        _values = list()
        while ancestor._ancestor is not None:
            _values.append(ancestor._value)
            ancestor = ancestor._ancestor
        _values.reverse()
        return _values

    def values(self):
        """Return all the lastest hour nodes.

        :return: latest hour nodes
        :rtype: list[float]
        """
        _all = list()
        for terminal_node in Possibilities.terminal_nodes:
            _all.append(terminal_node._val())
        return _all


class DayDynamicProgramming:
    """Provide a dynamic programing implementation of the optimization of actions depicted by StateModel."""

    class _HourNode:
        """Internal class for building a digraph where each node is standing for a state vector at time k and the actions that led to this node from the previous node (time k-1)."""

        NodeCounter = 0

        def __init__(self, state_vector: numpy.array, global_hour_index: int, estimated_outputs: numpy.array,ancestor_node = None, leading_actions=None, preference_value = 0):
            """Initialize a node standing for an possible state vector at a specific hour.

            :param state_vector: possible state vector
            :type state_vector: numpy.array
            :param global_hour_index: hour index
            :type global_hour_index: int
            :param ancestor_node: node of the previous hour that led to this node, defaults to None
            :type ancestor_node: _HourNode, optional
            :param leading_actions: actions (name and value) that led to this node, defaults to None
            :type leading_actions: dict[str,float], optional
            :param preference_value: assessment of the current state, defaults to 0
            :type preference_value: int, optional
            """
            self.ancestor_node = ancestor_node


            self._node_id = DayDynamicProgramming._HourNode.NodeCounter
            self.estimated_outputs = estimated_outputs
            self.global_hour_index = global_hour_index
            self.state_vector = state_vector
            self.discrete_state_vector = self._discretize_state(state_vector)
            self._preference_value_km1 = preference_value
            self.leading_actions_km1 = leading_actions
            DayDynamicProgramming._HourNode.NodeCounter += 1

        @property
        def cumulated_preference_value_km1(self) -> float:
            """Return the sum of the preference values from the first hour and according to the bloodline.

            :return: sum of preference values
            :rtype: float
            """
            _cumulated_preference_value_km1 = self._preference_value_km1
            if self.ancestor_node is not None:
                _cumulated_preference_value_km1 += self.ancestor_node.cumulated_preference_value_km1
            return _cumulated_preference_value_km1

        @property
        def node_id(self) -> int:
            """Return the id of the node.

            :return: unique identifier
            :rtype: int
            """
            return self._node_id

        def _discretize_state(self, state_vector: numpy.array):
            """Return a discrete-value representation of the state vector corresponding to the specified state space resolution.

            :param state_vector: the state vector in the continuous space
            :type state_vector: numpy.array
            :return: pair of indices referencing the cell of the virtual grid to which the state of the node belongs
            :rtype: tuple[int]
            """
            _discrete_state_vector = list()
            for state_variable_index in range(len(state_vector)):
                _discrete_state_vector.append(int(round(state_vector[state_variable_index, 0] / DayDynamicProgramming._HourNode.state_resolutions[state_variable_index])))
            return _discrete_state_vector

        def __cmp__(self, other_node: 'DayDynamicProgramming._HourNode') -> int:
            """Define comparison between nodes according to virtual state grid indices. Priority is given to first indices.

            :param other_node: other node
            :return: -1 if cell indices of current node <= other_node cell indices, 1 if greater than and 0 in case of equality
            """
            for i in range(len(self.discrete_state_vector)):
                if self.discrete_state_vector[i] < other_node.discrete_state_vector[i]:
                    return -1
                elif self.discrete_state_vector[i] > other_node.discrete_state_vector[i]:
                    return 1
            return 0

        def __lt__(self, other_node: 'DayDynamicProgramming._HourNode') -> bool:
            """Redefine 'lower than' legacy method.

            :param other_node: other node
            :return: True if node state grid cell indices < other_node state grid cell indices, False otherwise
            """
            return self.__cmp__(other_node) == -1

        def __eq__(self, other_node: 'DayDynamicProgramming._HourNode') -> bool:
            """Redefine 'equal to' legacy method.

            :param other_node: other node
            :return: True if node state grid cell indices == other_node state grid cell indices, False otherwise
            """
            return self.__cmp__(other_node) == 0

        def __gt__(self, other_node: 'DayDynamicProgramming._HourNode') -> bool:
            """Redefine 'greater than' legacy method.

            :param other_node: other node
            :return: True if node state grid cell indices > other_node state grid cell indices, False otherwise
            """
            return self.__cmp__(other_node) == 1

        def __str__(self) -> str:
            """Return a descriptive string of the node.

            :return: description of the node
            :rtype: str
            """
            string = 'N:' + str(self.node_id) + '(h:' + str(self.hour_in_day) + ', v:' + str(self.cumulated_preference_value_km1) + ', a:'
            if self.leading_actions_km1 is not None:
                for k,v in self.leading_actions_km1.items():
                    string += '%s > %i, ' % (k, v)
            if self.ancestor_node is not None:
                string += ' > ' + str(self.ancestor_node.node_id)
            return string

    def __init__(self, a_model: buildingenergy.model.Model,  initial_state_vector: numpy.array, global_starting_hour_index: int, state_resolutions, preference: buildingenergy.model.Preference):
        """Intialize the dynamic programing implementation for the optimization of actions. The optimization problem is related to times from hour_index till hour_index + 24 hour steps.

        :param model: model of the building
        :type model.H358Model, subclassing Model
        :param initial_state_vector: class containing the state model
        :type initial_state_vector: numpy.array
        :param global_starting_hour_index: define the size of the state space grid for selecting best solutions in each cell.
        :type global_starting_hour_index: int
        :param state_resolutions: resolution related to the state vector to reduce the number of propagated state vectors at each time step. Same size than the state vector.
        :type state_resolutions: list[float]
        :param preference: occupant preference
        :type preference: buildingenergy.model.Preference
        """
        print('# dynamic programming')
        print('## date: ', a_model.data('stringtime')[global_starting_hour_index])
        DayDynamicProgramming._HourNode.state_resolutions = state_resolutions
        self.model = a_model

        U0, _ = a_model.computeU(0, initial_state_vector, None, None)
        Y0 = a_model.computeY(0, initial_state_vector, U0, None)
        self.day_hour_node_lists = [[DayDynamicProgramming._HourNode(initial_state_vector, global_starting_hour_index, Y0)]]
        for k in range(global_starting_hour_index + 1, global_starting_hour_index + 25):  # k refers to the nodes under creation
            hour_nodes_list_for_hour_k = []  # contains the nodes of the current hour k
            action_names, possible_actions = self.generate_possible_actions_at_given_hour(k-1)  # generate the possible actions for hour k
            leading_action_possibilities = Possibilities(possible_actions).values()   #generate a list with combinations of possible action values, respecting the order in action_names
            for leading_action_possibility in leading_action_possibilities:
                leading_action_km1 = dict(zip(action_names, leading_action_possibility))  # create a dict with action_names as keys and corresponding values (leading_action_possibility) as values: 1 action name is related to 1 action value
                for hour_node_km1 in self.day_hour_node_lists[-1]:  # select all the nodes of the preceding hour
                    influencing_variable_km1 = (leading_action_km1['door_opening'], leading_action_km1['window_opening'])
                    Ukm1, leading_action_km1 = a_model.computeU(k-1, hour_node_km1.state_vector, influencing_variable_km1, leading_action_km1)
                    currentY_km1 = a_model.computeY(k-1, hour_node_km1.state_vector, Ukm1, influencing_variable_km1)
                    state_vector_k, _ = a_model.stepX(k-1, hour_node_km1.state_vector, None, influencing_variable_km1, Ukm1)

                    presence=self.model.data('occupancy')[k-1]
                    if k < len(self.model.data('occupancy')):
                        presence = presence + self.model.data('occupancy')[k]
                    preference_value_km1 = preference.assess(leading_action_km1['heating_power'], currentY_km1[0,0], currentY_km1[1,0], presence)
                    new_node = DayDynamicProgramming._HourNode(state_vector_k, k, currentY_km1, hour_node_km1, leading_action_km1, preference_value_km1)
                    hour_nodes_list_for_hour_k.append(new_node)
            print('hour %i: %i new nodes' % (k, len(hour_nodes_list_for_hour_k)), end=' ')
            hour_nodes_list_for_hour_k.sort()
            filtered_next_hour_node_list = list()
            current_cell, current_vals, current_nodes = [], [], []
            for k in range(len(hour_nodes_list_for_hour_k)):
                if len(current_cell) == 0 or current_cell[-1] == hour_nodes_list_for_hour_k[k].discrete_state_vector:
                    current_cell.append(hour_nodes_list_for_hour_k[k].discrete_state_vector)
                    current_vals.append(hour_nodes_list_for_hour_k[k].cumulated_preference_value_km1)
                    current_nodes.append(hour_nodes_list_for_hour_k[k])
                else:
                    i_best = current_vals.index(min(current_vals))
                    filtered_next_hour_node_list.append(current_nodes[i_best])
                    current_cell, current_vals, current_nodes = [], [], []
            if len(current_cell) > 0:
                i_best = current_vals.index(min(current_vals))
                filtered_next_hour_node_list.append(current_nodes[i_best])

            print('-> remaining %i new nodes' % (len(filtered_next_hour_node_list)))
            self.day_hour_node_lists.append(filtered_next_hour_node_list)
        terminal_nodes = self.day_hour_node_lists[-1]
        terminal_nodes_indices_values = dict()
        for i in range(len(terminal_nodes)):
            terminal_nodes_indices_values[i] = terminal_nodes[i].cumulated_preference_value_km1
        sorted_terminal_nodes_indices_values = [(i, v) for i, v in sorted(terminal_nodes_indices_values.items(), key=lambda item: item[1])]
        best_terminal_node = terminal_nodes[sorted_terminal_nodes_indices_values[0][0]]
        self._best_strategy = self._strategy(best_terminal_node)

        print("* 10 best strategies")
        counter = 0
        for i,v in sorted_terminal_nodes_indices_values:
            current_strategy = self._strategy(terminal_nodes[i])
            string = str(i) + ':'
            for k in range(len(current_strategy['door_opening'])):
                string += str(k)+'['
                for action in current_strategy:
                    if action == 'door_opening'and current_strategy[action][k]==1:
                        string += 'D'
                    elif action == 'window_opening'and current_strategy[action][k]==1:
                        string += 'W'
                    elif action == 'temperature_setpoint':
                        string += 'T' + str(current_strategy[action][k])
                    elif action == 'heating_power':
                        string += 'P%.0f' % current_strategy[action][k]
                string += ']'
            string += '>%f%%, ' % (100*v / len(current_strategy['door_opening']))
            if counter > 9:
                break
            counter += 1
            print(string + '\n')

    def _strategy(self, terminal_node):
        _results = {'door_opening': list(), 'window_opening': list(), 'heating_power': list(), 'temperature_setpoint': list()}
        estimated_outputs = None
        while terminal_node.ancestor_node is not None:
            resulting_actions = terminal_node.leading_actions_km1
            if estimated_outputs is None:
                estimated_outputs = terminal_node.estimated_outputs
            else:
                estimated_outputs = numpy.concatenate((terminal_node.estimated_outputs, estimated_outputs), axis=1)
            _results['door_opening'].insert(0, resulting_actions['door_opening'])
            _results['window_opening'].insert(0, resulting_actions['window_opening'])
            _results['heating_power'].insert(0, resulting_actions['heating_power'])
            if 'temperature_setpoint' in resulting_actions:
                _results['temperature_setpoint'].insert(0, resulting_actions['temperature_setpoint'])
            else:
                _results['temperature_setpoint'].insert(0, None)
            _results['estimated_outputs'] = estimated_outputs
            terminal_node = terminal_node.ancestor_node
        return _results

    @property
    def results(self):
        """Return results of the optimization.

        :return: list of actions with optimal values covering 24 hours
        :rtype: dict[str,list[float]]
        """
        return self._best_strategy

    def generate_possible_actions_at_given_hour(self, global_starting_hour_index: int):
        """Generate the possible action at hour k as a list of action names and a list of tuples with the possible values corresponding to the action with the same index in the list (1).

        :param global_starting_hour_index: hour index
        :type global_starting_hour_index: int
        :return: a list of names of possible actions and a list of tuples with possible action values
        :rtype: tuple[list[str], list[tuple(float)]]
        """
        raise NotImplemented



"""Time management helper.

stephane.ploix@g-scop.grenoble-inp.fr
"""

import datetime
import time

def epochtimems_to_stringdate(epochtimems, date_format='%d/%m/%Y %H:%M:%S'):
    """Transform an epoch time  into a string representation.

    :param epochtimems: epoch time in milliseconds
    :type epochtimems: int
    :return: string representation '%d/%m/%Y %H:%M:%S'
    :rtype: datetime.datetime
    """
    return time.strftime(date_format, time.localtime(epochtimems // 1000))


def epochtimems_to_datetime(epochtimems):
    """Transform an epoch time into an internal datetime representation.

    :param epochtimems: epoch time in milliseconds
    :type epochtimems: int
    :return: internal datetime representation
    :rtype: datetime.datetime
    """
    return datetime.datetime.fromtimestamp(epochtimems // 1000)


def datetime_to_epochtimems(a_datetime):
    """Transform a an internal datetime representation into a epoch time.

    :param a_datetime: internal datetime representation
    :type a_datetime: datetime to be converted
    :return: epoch time in milliseconds
    :rtype: int
    """
    return a_datetime.timestamp() * 1000


def stringdate_to_epochtimems(stringdatetime, date_format='%d/%m/%Y %H:%M:%S'):
    """Transform a date string representation into an epoch time.

    :param stringdatetime: date string representation '%d/%m/%Y %H:%M:%S'
    :type stringdatetime: str
    :return: epoch time in milliseconds
    :rtype: int
    """
    stringdatetime_secs_ms = stringdatetime.split(',')
    epochdatems = int(time.mktime(time.strptime(stringdatetime_secs_ms[0], date_format)) * 1000)
    if len(stringdatetime_secs_ms) > 1:
        epochdatems = epochdatems + int(stringdatetime_secs_ms[1])
    return epochdatems


def datetime_to_stringdate(a_datetime, date_format='%d/%m/%Y %H:%M:%S'):
    """Transform a datetime representation into a datetime internal format.

    :param a_datetime: internal datetime representation
    :type a_datetime: datetime.datetime
    :return: stringdatetime: date string representation '%d/%m/%Y %H:%M:%S'
    :rtype: str
    """
    return a_datetime.strftime(date_format)


def stringdate_to_datetime(stringdatetime, date_format='%d/%m/%Y %H:%M:%S'):
    """Transform a date string representation into an internal datetime representation.

    :param stringdatetime: date string representation '%d/%m/%Y %H:%M:%S'
    :type stringdatetime: str
    :return: internal datetime representation
    :rtype: datetime.datetime
    """
    return datetime.datetime.fromtimestamp(time.mktime(time.strptime(stringdatetime, date_format)))


def epochtimems_to_timequantum(epochtimems, timequantum_duration_in_secondes):
    """Transform an epoch time into a rounded discrete epoch time according to a given time quantum (sampling period).

    :param epochtimems: epoch time in milliseconds
    :type epochtimems: int
    :param timequantum_duration_in_secondes: time quantum duration (sampling period) in seconds
    :type timequantum_duration_in_secondes: int
    :return: rounded discrete epoch time in milliseconds
    """
    return (epochtimems // (timequantum_duration_in_secondes * 1000)) * timequantum_duration_in_secondes * 1000


def get_stringdate_with_day_delta(numberofdays=0, date_format='%d/%m/%Y %H:%M:%S'):
    """Compute a date from today minus a given day delta.

    :param numberofdays: number of day to remove to the current date, defaults to 0
    :type numberofdays: int, optional
    :param date_format: date format, defaults to '%d/%m/%Y %H:%M:%S'
    :type date_format: str, optional
    :return: the date in the past
    :rtype: datetime.datetime
    """
    return (datetime.datetime.now() - datetime.timedelta(days=numberofdays)).strftime(date_format)


def current_stringdate(date_format='%d/%m/%Y %H:%M:%S'):
    """Return the current date in string format.

    :param date_format: the string format, defaults to '%d/%m/%Y %H:%M:%S'
    :type date_format: str, optional
    :return: current date in string
    :rtype: str
    """
    return time.strftime(date_format, time.localtime())


def current_epochtimems():
    """Return current date in epoch time format.

    :return: epoch time number of miliseconds
    :rtype: int
    """
    return int(time.mktime(time.localtime()) * 1000)



"""A module that contains a data container of on-site measurements.

Author:  stephane.ploix@g-scop.grenoble-inp.fr
"""
import pandas
import tkinter as tk
from buildingenergy import timemg
import matplotlib
import matplotlib.pyplot
import matplotlib.dates
matplotlib.use('TkAgg')


class DataContainer:
    """It gathers measurement data corresponding to dates with a regular sample time. A DataContainer gives an easy access to them and make it possible to plot them."""

    def __init__(self, csv_filename, skiprows=0, nrows=None, initial_string_date=None, final_string_date=None):
        """Create a data container by collecting data from a csv file.

        Selection of a time period can be achieved in 2 ways: with a row basis using skiprows and nrows, or with a date basis: the intersection of both periods is returned.

        :param csv_filename: name of the csv file containing
        :type csv_filename: str
        :param skiprows: number of rows to ignore at the begining of the csv file, defaults to 0
        :type skiprows: int, optional
        :param nrows: number of rows to read (from skiprows), defaults to None
        :type nrows: int, optional
        :param initial_string_date: initial date in format 'dd/mm/YYYY HH:MM:SS', default to None, optional
        :type initial_string_date: str
        :param final_string_date: final date in format 'dd/mm/YYYY HH:MM:SS', default to None, optional
        :type final_string_date: str
        """
        self.sample_time = None
        self.starting_stringdatetime = None
        self.ending_stringdatetime = None
        self.registered_databases = dict()
        self.data = dict()
        self.extracted_variables = list()
        self.data['stringtime'] = None
        self.extracted_variables.append('stringtime')
        self.data['epochtime'] = None
        self.extracted_variables.append('epochtime')
        self.data['datetime'] = None
        self.extracted_variables.append('datetime')

        self.SPECIAL_VARIABLES = ('stringtime', 'epochtime', 'datetime')
        self.variable_full_name_id_dict = dict()  # context$variable: variable_id
        self.variable_full_name_database_dict = dict()  # context$variable: database_name
        self.variable_full_name_csv_dict = dict()  # context$variable: csv_file_name
        self.variable_type = dict()
        self.contexts = list()

        self.data = dict()
        self._extracted_variable_full_names = list()  # context_variable names
        self.data['stringtime'] = None
        self._extracted_variable_full_names.append('stringtime')
        self.data['epochtime'] = None
        self._extracted_variable_full_names.append('epochtime')
        self.data['datetime'] = None
        self._extracted_variable_full_names.append('datetime')

        dataframe = pandas.read_csv(csv_filename, dialect='excel')

        variable_names = dataframe.columns
        for variable_name in variable_names:
            if variable_name == 'stringtime':
                dataframe['stringtime'].astype({'stringtime': 'str'})
                self.data[variable_name] = dataframe[variable_name].values.tolist()[skiprows:nrows+skiprows] if nrows is not None else dataframe[variable_name].values.tolist()[skiprows:]
            elif variable_name == 'datetime':
                if nrows is not None:
                    self.data[variable_name] = [timemg.stringdate_to_datetime(stringdatetime, date_format='%Y-%m-%d %H:%M:%S') for stringdatetime in dataframe['datetime'].values.tolist()[skiprows:nrows+skiprows]]
                else:
                    self.data[variable_name] = [timemg.stringdate_to_datetime(stringdatetime, date_format='%Y-%m-%d %H:%M:%S') for stringdatetime in dataframe['datetime'].values.tolist()[skiprows:]]
            elif variable_name == 'epochtime':
                dataframe[variable_name].astype({'epochtime': 'int'})
                self.data[variable_name] = dataframe[variable_name].values.tolist()[skiprows:nrows+skiprows] if nrows is not None else dataframe[variable_name].values.tolist()[skiprows:]
            else:
                self.add_external_variable(variable_name, dataframe[variable_name].values.tolist()[skiprows:nrows+skiprows] if nrows is not None else dataframe[variable_name].values.tolist()[skiprows:])

        in_period=False
        k_inf, k_sup = 0, len(self.data['datetime'])
        initial_datetime = timemg.stringdate_to_datetime(initial_string_date, date_format='%d/%m/%Y %H:%M:%S') if initial_string_date is not None else None
        final_datetime = timemg.stringdate_to_datetime(final_string_date, date_format='%d/%m/%Y %H:%M:%S') if final_string_date is not None else None
        for k in range(len(self.data['datetime'])):
            current_datetime = self.data['datetime'][k]
            if not in_period and initial_datetime is not None and current_datetime >= initial_datetime:
                k_inf = k
                in_period = True
            if in_period and final_datetime is not None and current_datetime >= final_datetime:
                k_sup = k
                break
        for variable_name in variable_names:
            self.data[variable_name] = self.data[variable_name][k_inf:k_sup]

        self.starting_stringdatetime = self.data['stringtime'][0]
        self.ending_stringdatetime = self.data['stringtime'][-1]
        self.sample_time_in_secs = int((self.data['epochtime'][1]-self.data['epochtime'][0]) / 1000)

    def add_external_variable(self, label: str, values):
        """Use to add a series of values to the container. It will appears as any other measurements.

        :param label: file_name of the series
        :type label: str
        :param values: series of values but it must be compatible with the times which are common to all series
        :type values: list[float]
        """
        if label in self.extracted_variables:
            print('variable %s replaced because it was already extracted' % label)
            del self.data[label]
        self.data[label] = values
        self.extracted_variables.append(label)

    def get_variable(self, label):
        """Use to get a series of values. Special series of dates are available: 'stringtime' as string, 'epochtime' as number of seconds from January 1st 1970 at Greenwich (EPOCH time),  'datetime' as Python date format.

        :param label: file_name of the series coming from CSV file or among 'stringtime', 'epochtime' or 'datetime'
        :type label: str
        :return: a list of floating numbers
        :rtype: list[float]
        """
        return self.data[label]

    def get_number_of_variables(self):
        """Get the number of series including 3 times series named 'stringtime', 'epochtime' or 'datetime'.

        :return: number of series plus 3 different time series
        :rtype: int
        """
        return len(self.extracted_variables)

    def get_number_of_samples(self):
        """Return the number of data common to all the series.

        :return: the number of data common to all the series
        :rtype: int
        """
        if self.data['epochtime'] is None:
            return 0
        else:
            return len(self.data['epochtime'])

    def _plot_selection(self, int_vars: list):
        """Use to plot curve.

        :param int_vars: reference to the variable to be plotted
        :type: list[int]
        """
        styles = ('-', '--', '-.', ':')
        linewidths = (3.0, 2.5, 2.5, 1.5, 1.0, 0.5, 0.25)
        figure, axes = matplotlib.pyplot.subplots()
        axes.set_title('from %s to %s' % (self.starting_stringdatetime, self.ending_stringdatetime))
        text_legends = list()
        for i in range(len(int_vars)):
            if int_vars[i].get():
                style = styles[i % len(styles)]
                linewidth = linewidths[i // len(styles) % len(linewidths)]
                time_data = list(self.data['datetime'])
                value_data = list(self.data[self.extracted_variables[i + 3]])
                if len(time_data) > 1:
                    time_data.append(time_data[-1] + (time_data[-1] - time_data[-2]))
                    value_data.append(value_data[-1])
                axes.step(time_data, value_data, linewidth=linewidth, linestyle=style, where='post')
                axes.set_xlim([time_data[0], time_data[-1]])
                text_legends.append(self.extracted_variables[i + 3])
                int_vars[i].set(0)
        axes.legend(text_legends, loc=0)
        figure.set_tight_layout(True)
        axes.xaxis.set_major_locator(matplotlib.dates.WeekdayLocator(byweekday=matplotlib.dates.MO))
        axes.xaxis.set_major_formatter(matplotlib.dates.DateFormatter('%d/%m %H:%M'))
        axes.xaxis.set_minor_locator(matplotlib.dates.DayLocator())
        axes.grid(True)
        matplotlib.pyplot.show()

    def plot(self):
        """Display a series selector to plot the different time series."""
        tk_variables = list()
        tk_window = tk.Tk()
        tk_window.wm_title('variable plotter')
        tk.Button(tk_window, text='plot', command=lambda: self._plot_selection(tk_variables)).grid(row=0, column=0, sticky=tk.W + tk.E)
        frame = tk.Frame(tk_window).grid(row=1, column=0, sticky=tk.N + tk.S)
        vertical_scrollbar = tk.Scrollbar(frame, orient=tk.VERTICAL)
        vertical_scrollbar.grid(row=1, column=1, sticky=tk.N + tk.S)
        canvas = tk.Canvas(frame, width=400, yscrollcommand=vertical_scrollbar.set)
        tk_window.grid_rowconfigure(1, weight=1)
        canvas.grid(row=1, column=0, sticky='news')
        vertical_scrollbar.config(command=canvas.yview)
        checkboxes_frame = tk.Frame(canvas)
        checkboxes_frame.rowconfigure(1, weight=1)
        for i in range(3, len(self.extracted_variables)):
            tk_variable = tk.IntVar()
            tk_variables.append(tk_variable)
            tk.Checkbutton(checkboxes_frame, text=self.extracted_variables[i], variable=tk_variable, offvalue=0).grid(row=(i - 3), sticky=tk.W)
        canvas.create_window(0, 0, window=checkboxes_frame)
        checkboxes_frame.update_idletasks()
        canvas.config(scrollregion=canvas.bbox('all'))
        tk_window.geometry(str(tk_window.winfo_width()) + "x" + str(tk_window.winfo_screenheight()))
        tk_window.mainloop()

    def size(self):
        """Return the recorded date size.

        :return: the data size
        :rtype: int
        """
        return len(self.data['stringtime'])

    def __str__(self):
        """Make it possible to print a datacontainer.

        :return: description of the datacontainer
        :rtype: str
        """
        string = 'Data cover period from %s to %s with time period: %d seconds\nRegistered database:\n' % (self.starting_stringdatetime, self.ending_stringdatetime, self.sample_time)
        for database in self.registered_databases:
            string += '- %s \n' % database
        string += 'Available variables:\n'
        for variable_name in self.extracted_variables:
            string += '- %s \n' % variable_name
        return string


class PlotSaver:
    """Class to save plots into file."""

    def __init__(self, data_container: DataContainer):
        """Initialize the plot saver.

        :param data_container: datacontainer from which time series can be plotted
        :type data_container: DataContainer
        """
        self.data_container = data_container

    def time_plot(self, selected_variable_full_names, filename: str):
        """Generate time plot.

        :param selected_variable_full_names: full names (ie context$name) of the variables to be plot
        :type selected_variable_full_names: list[str]
        :param filename: name of the image file that will contain the plot
        :type filename: str
        """
        styles = ('-', '--', '-.', ':')
        linewidths = (1.0, 1.0, 0.75, 0.75, 0.5, 0.5, 0.25, 0.25)
        _, axes = matplotlib.pyplot.subplots()
        text_legends = list()
        for i in range(len(selected_variable_full_names)):
            style = styles[i % len(styles)]
            linewidth = linewidths[i // len(styles) % len(linewidths)]
            time_data = list(self.data_container.data['datetime'])
            variable_data = list(self.data_container.get_variable(selected_variable_full_names[i]))
            if len(time_data) > 1:
                time_data.append(time_data[-1] + (time_data[-1] - time_data[-2]))
                variable_data.append(variable_data[-1])
            axes.step(time_data, variable_data, linewidth=linewidth, linestyle=style, where='post')
            axes.set_xlim([time_data[0], time_data[-1]])
            text_legends.append(selected_variable_full_names[i])

        # axes.xaxis.set_major_locator(matplotlib.dates.WeekdayLocator(byweekday=matplotlib.dates.MO))
        # axes.xaxis.set_major_formatter(matplotlib.dates.DateFormatter('%d/%m %H:%M'))
        axes.legend(text_legends, loc=0)
        axes.xaxis.set_minor_locator(matplotlib.dates.DayLocator())
        axes.fmt_xdata = matplotlib.dates.DateFormatter('%d/%m/%Y %H:%M')
        matplotlib.pyplot.gcf().autofmt_xdate()
        axes.grid(True)
        matplotlib.pyplot.savefig(filename+'.png', dpi=1500)
